/*
 * Asterisk -- A telephony toolkit for Linux.
 * Copyright (C) 1999-2005, Digium, Inc.
 * Mark Spencer <markster@digium.com>  - Asterisk main author
 *
 * app_dialipcontact.c: Dial IPContact applications
 *
 * Copyright (c) 2004-2008, IPContact Software SRL.  All rights reserved.
 * Authors:
 *
 * Luis Vazquez <luis@ipcontact.com.uy>
 * Daniel Ferrer <daniel@ipcontact.com.uy>
 *
 * $Id: app_dialipcontact.c,v 1.31 2007-11-19 15:05:45 ipcontact Exp $
 *
 * This code is released by the author with no restrictions on usage.
 *
 */


/*! \file
 *
 * \brief IPContact Dialing applications
 *
 * \author Luis Vazquez <luis@ipcontact.com.uy>
 * \author Daniel Ferrer <daniel@ipcontact.com.uy>
 * 
 * \ingroup applications
 */


/*** MODULEINFO
		 <depend>mysqlclient</depend>
		 <depend>res_ipcontact</depend>
***/

#include <asterisk.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <unistd.h>
#include <ctype.h>
#include <string.h>
#include <sys/socket.h>

#include <asterisk/logger.h>
#include <asterisk/options.h>
#include <asterisk/channel.h>
#include <asterisk/pbx.h>
#include <asterisk/utils.h>
#include <asterisk/callerid.h>
#include <asterisk/module.h>
#include <asterisk/config.h>
#include <asterisk/astdb.h>
#include <asterisk/indications.h>
#include <asterisk/manager.h>
#include <asterisk/say.h>
#include <asterisk/app.h>
#include <time.h>

//#include <pcre.h>


#include "res_ipcontact.h"


#define CALL_FORWARD_BASE "ipcontact/callforward"
#define CALL_FORWARD_ALL "all"
#define CALL_FORWARD_BUSY "busy"
#define CALL_FORWARD_UNAVAIL "unavail"

#define MODULE_DESC "IPContact Dialing Applications"

#ifdef HAS_XMLDOC
/*** DOCUMENTATION
	<application name="DialHuntGroup" language="en_US">
		<synopsis>
			Dial a number over given hunt group
		</synopsis>
		<syntax>
			<parameter name="huntgroup" required="true">
				<para>HuntGroup ID to use.</para>
			</parameter>
			<parameter name="number" required="true">
				<para>Phone Number to dial on specified huntgroup ID.</para>
			</parameter>
			<parameter name="options" required="false">
				<optionlist>
					<option name="1">
						<para><emphasis>Fail on Busy</emphasis>, hunting gives up when Dial app returns BUSY condition (default behaviour
						if no option is specifed). For compatibility with IPContact Addons 1.4.4.</para>
					</option>
					<option name="c">
						<para><emphasis>Continue</emphasis>, ie not Fail on Busy. Hunting continues until there is no more lines.</para>
					</option>
					<option name="0">
						<para>Idem as <literal>c</literal>. For compatibility with 1.4.4.</para>
					</option>
					<option name="g">
					<para>Select the lowerest-id line in huntgroup (<emphasis>ascending sequential hunting</emphasis>). This is the default if none option specified.</para>
					</option>
					<option name="G">
						<para>Select the highest-id line in huntgroup (<emphasis>descending sequential hunting</emphasis>).</para>
				  	</option>
					<option name="r">
						<para>Use a round robin search, starting at next highest-id line than last time (<emphasis>ascending rotary hunting</emphasis>).</para>
					</option>
					<option name="R">
						<para>Use a round robin search, starting at next lowest-id line than last time (<emphasis>descending rotary hunting</emphasis>).</para>
				    </option>
			    </optionlist>
			</parameter>
			<parameter name="dialopts" required="false">
				<para>Extra arguments to Dial app</para>
			</parameter>
		</syntax>
		<description>
				<para>Find a free line and dial the passed number on given Hunt Group from MySQL DB.</para>
		</description>
	</application>

	<application name="SaveClientCallInfo" language="en_US">
		<synopsis>
			  Writes client call info to MySQL database
		</synopsis>
		<syntax>
		</syntax>
		<description>
				<para>Writes caller and called ID info into MySQL database. This app generates events based on standard channel variables and also the following vars:</para>
				<para><variable>INCOMING_CALL</variable>, <variable>OUTGOING_CALL</variable>, <variable>CALLEDNUMBER</variable>, <variable>CLIENT_DN</variable>,</para>
				<para><variable>CLIENT_CN</variable>, <variable>AGENT_NAME</variable>, <variable>GROUP_ID</variable>.</para>
		</description>
	</application>

	<application name="VirtualPBXGoto" language="en_US">
		<synopsis>
				Goto Virtual PBX specific section mainmenu, operator, voicemail
		</synopsis>
		<syntax>
			<parameter name="location" required="true">
			    <enumlist>
					<enum name="m">
						<para>Mainmenu</para>
					</enum>
					<enum name="o">
						<para>Operator</para>
					</enum>
					<enum name="v">
						<para>Voicemail</para>
					</enum>
			    </enumlist>
	        </parameter>
	    </syntax>
	    <description>
		    <para>Jumps to specified context,extension,priority on VPBX table according to variable <variable>VPBX</variable></para>
	    </description>
		<see-also>
			   CLI command <literal>ipcontact show vpbxs</literal>
		</see-also>
	</application>

	<application name="GetPinProfile" language="en_US">
		<synopsis>
			  Get auth information for a user, given a pin number.
		</synopsis>
		<syntax>
			<parameter name="pin" required="true">
				<para>User's pin</para>
			</parameter>
		</syntax>
		<description>
				<para>Get authentication information from user, associated with a given pin number. Sets the following variables on return:</para>
				<para><variable>AUTH_user</variable>, <variable>AUTH_local</variable>, <variable>AUTH_longdist</variable>, <variable>AUTH_cellular</variable>,</para>
				<para><variable>AUTH_inter</variable>.</para>
		</description>
	</application>

	<application name="GetExtenProfile" language="en_US">
		<synopsis>
			  Get auth information given an extension number.
		</synopsis>
		<syntax>
			<parameter name="exten" required="true">
				<para>Extension number</para>
			</parameter>
		</syntax>
		<description>
				<para>Get authentication information for user, associated with given pin number. Sets the following variables on return:</para>
				<para><variable>AUTH_local</variable>, <variable>AUTH_longdist</variable>, <variable>AUTH_cellular</variable>, <variable>AUTH_inter</variable>.</para>
		</description>
	</application>

	<application name="CallForward" language="en_US">
		<synopsis>
			  Set CallForward status for given extension
		</synopsis>
		<syntax>
			<parameter name="type" required="true">
			    <enumlist>
					<enum name="a">
						<para>All</para>
					</enum>
					<enum name="b">
						<para>Busy</para>
					</enum>
					<enum name="u">
						<para>Unavailable</para>
					</enum>
				</enumlist>
			</parameter>
			<parameter name="status" required="true">
			    <enumlist>
					<enum name="on">
						<para>Enable CF</para>
					</enum>
					<enum name="off">
						<para>Disable CF</para>
					</enum>
				</enumlist>
			</parameter>
		</syntax>
		<description>
				<para>Enable or disable call forwarding of given type for current extension</para>
		</description>
	</application>
	
	<application name="SetCIDFromExten" language="en_US">
		<synopsis>
				Set CallerID as user's extension on SIP calls
		</synopsis>
		<syntax>
		</syntax>
		<description>
			 <para>If the caller is a SIP channel reads the current value of cid number and if this is not a valid extension number uses the user name in the From of the SIP message as a key to search the extension in the IPContact DB. If found, this extension number is set as the new cid number</para>
		</description>
	</application>

***/
#endif

static char *app_dialhunt = "DialHuntGroup";
#ifndef HAS_XMLDOC
static char *synopsis_dialhunt = "Dial a number over given hunt group";
static char *descrip_dialhunt =
	"  DialHuntGroup(hunt" AST_STANDARD_APP_DELIM "number" AST_STANDARD_APP_DELIM "options" AST_STANDARD_APP_DELIM "dialopts):\n"
	"Find a free line and dial the passed number on given Hunt Group from MySQL DB.\n\n"
  "Options:\n"
  "  1 - Fail on Busy, hunting gives up when Dial app returns BUSY condition (default behaviour\n"
  "      if no option is specifed). For compatibility with IPContact Addons 1.4.4.\n"
  "  c - Continue, ie 'not Fail on Busy'. Hunting continues until there is no more lines.\n"
  "  0 - Idem as 'c'. For compatibility with 1.4.4.\n"
  " Options to control hunting order\n"
  "  g - Select the lowerest-id line in huntgroup (ascending sequential hunting).\n"
  "      This is the default if none option specified.\n"
  "  G - Select the highest-id line in huntgroup. (descending sequential hunting).\n"
  "  r - Use a round robin search, starting at next highest-id line than last time (ascending rotary hunting).\n"
  "  R - Use a round robin search, starting at next lowest-id line than last time (descending rotary hunting).\n"
	"Returns whatever value the dial returns 0 or -1 when a line cannot be found.\n";
#endif

#ifdef HAS_AST18
static int dialhunt_exec(struct ast_channel *chan, const char *data);
#else
static int dialhunt_exec(struct ast_channel *chan, void *data);
#endif

static char *app_setcidext = "SetCIDFromExten";
#ifndef HAS_XMLDOC
static char *synopsis_setcidext = "Set CallerID as user's extension on SIP calls";
static char *descrip_setcidext =
	"SetCIDFromExten()\n"
	"  If the caller is a SIP channel reads the current value of cid number\n"
	"and if this is not a valid extension number uses the user name in the From\n"
	"of the SIP message as a key to search the extension in the IPContact DB.\n"
	"If found, this extension number is set as the new cid number.\n";
#endif
#ifdef HAS_AST18
static int setcidext_exec(struct ast_channel *chan, const char *data);
#else
static int setcidext_exec(struct ast_channel *chan, void *data);
#endif

static char *app_callforward = "CallForward";
#ifndef HAS_XMLDOC
static char *synopsis_callforward = "Enable/Disable Call Forward on dialing extension";
static char *descrip_callforward =
	"CallForward(type[" AST_STANDARD_APP_DELIM "<on|off>])\n"
	"  Enable or disable call forwarding of given type (a:all, b:busy, u:unavail) on current extension.\n";
#endif
#ifdef HAS_AST18
static int callforward_exec(struct ast_channel *chan, const char*data);
#else
static int callforward_exec(struct ast_channel *chan, void *data);
#endif

static char *app_savecallinfo = "SaveClientCallInfo";
#ifndef HAS_XMLDOC
static char *synopsis_savecallinfo = "Writes client call info to MySQL database";
static char *descrip_savecallinfo =
	"SaveClientCallInfo()\n" "  Writes caller and called ID info into MySQL database.\n"
	"  This app generates events based on standard channel variables and also the following vars:\n "
    "INCOMING_CALL, OUTGOING_CALL, CALLEDNUMBER, CLIENT_DN, CLIENT_CN, AGENT_NAME, GROUP_ID.\n";
#endif
/* Applicacion SaveClientCallInfo
   Para determinar el destino de las llamadas debe
   estar definida la variable OUTGOING_CALL en llamadas salientes */
#ifdef HAS_AST18
static int savecallinfo_exec(struct ast_channel *chan, const char *data);
#else
static int savecallinfo_exec(struct ast_channel *chan, void *data);
#endif

static char *app_vpbxgoto = "VirtualPBXGoto";
#ifndef HAS_XMLDOC
static char *synopsis_vpbxgoto = "VirtualPBXGoto([m|o|v])";
static char *descrip_vpbxgoto =
	"VirtualPBXGoto([m|o|v])\n"
	"  Go to Virtual PBX specific section:\n"
	"\t m => mainmenu, o => operator, v => voicemail.\n";
#endif
#ifdef HAS_AST18
static int vpbxgoto_exec(struct ast_channel *chan, const char *data);
#else
static int vpbxgoto_exec(struct ast_channel *chan, void *data);
#endif

static char *app_get_pin_profile = "GetPinProfile";
#ifndef HAS_XMLDOC
static char *synopsis_get_pin_profile = "GetPinProfile(pin)";
static char *descrip_get_pin_profile =
	"GetPinProfile(pin)\n"
	"  Get authentication for user associated with given pin number.\n"
	"Affected variables are: AUTH_user, AUTH_local, AUTH_longdist, AUTH_cellular, AUTH_inter.\n";
#endif
#ifdef HAS_AST18
static int get_pin_profile_exec(struct ast_channel *chan, const char *data);
#else
static int get_pin_profile_exec(struct ast_channel *chan, void *data);
#endif

static char *app_get_exten_profile = "GetExtenProfile";
#ifndef HAS_XMLDOC
static char *synopsis_get_exten_profile = "GetExtenProfile(exten)";
static char *descrip_get_exten_profile =
	"GetExtenProfile(exten)\n"
	"  Get authentication for an extension.\n"
	"Affected variables are: AUTH_local, AUTH_longdist, AUTH_cellular, AUTH_inter.\n";
#endif
#ifdef HAS_AST18
static int get_exten_profile_exec(struct ast_channel *chan, const char *data);
#else
static int get_exten_profile_exec(struct ast_channel *chan, void *data);
#endif

static const int first_digit_timeout = 5000;
static const int digit_timeout = 2000;
static char *indication_confirm_tone = "425/100,0/200,425/100,0/200,425/100,0/1000";

#define NO_CID_NAME ""

/* Globals */
static int debug = 0;

static const char config_file[] = "ipcontact/dialipcontact.conf";

extern int schemaversion;

static struct callinfo_config {
	int usedb;					/* Whether to save call info in SQL DB */
	int usemanager;				/* Whether to raise call info events to manager */
} callinfo = {
	.usedb = 0,
	.usemanager = 1,
};

static struct dialhunt_config {
	int failonbusy;				/* Whether to keep trying when dial returns BUSY */
	char dialopts[32];			/* Options to pass to Dial application */
} dialhunt = {
	.failonbusy = 1, 
	.dialopts = "",
};

enum {
	HUNT_FAILONBUSY =   (1 << 0),
	HUNT_NOFAILONBUSY = (1 << 2),
	HUNT_DIAL_SEQASC  = (1 << 3),
	HUNT_DIAL_SEQDESC = (1 << 4),
	HUNT_DIAL_ROTASC  = (1 << 5),
	HUNT_DIAL_ROTDESC = (1 << 6),
} dialhunt_exec_opts_args;

AST_APP_OPTIONS(dialhunt_exec_opts, {
	AST_APP_OPTION('1', HUNT_FAILONBUSY),
	AST_APP_OPTION('0', HUNT_NOFAILONBUSY),
	AST_APP_OPTION('c', HUNT_NOFAILONBUSY),
	AST_APP_OPTION('g', HUNT_DIAL_SEQASC),
	AST_APP_OPTION('G', HUNT_DIAL_SEQDESC),
	AST_APP_OPTION('r', HUNT_DIAL_ROTASC),
  AST_APP_OPTION('R', HUNT_DIAL_ROTDESC),
});

static struct ast_app *appDial = NULL;

enum call_direction { INCOMING_CALL = 0, LOCAL_CALL, OUTGOING_CALL };

struct line_info {
	char number[8];
	char protocol[16];
	char prefix[32];
	char postfix[32];
};

struct dial_info {
	int vpbx;
	char exten[32];
	char protocol[16];
	char channel[32];
	char dialstring[64];
	char vm[16];
	int vmflag;
	int monitor;
};

struct exten_profile {
	char exten[32];
	unsigned char local;
	unsigned char national;
	unsigned char cellular;
	unsigned char inter;
};

/**** Utility functions ****/
inline static int indication_comfirm(struct ast_channel *chan);
inline static int playback_file(struct ast_channel *, const char *);
inline static int build_dial_string(struct line_info *line, const char *number, char *dialstr, const int len);
/* fix_sip_cidnumber:
   Recibe un canal SIP y en caso de que el cid number no sea una extension valida
   busca si el canal sip esta asociado a alguna extension en la base de datos.
   En caso de encontrarla setea este valor como nuevo cid number y crea el nuevo
   string del callerID en el buffer que se pasa como parmetro.
   Devuelve 0 si se pudo modificar el cid number por una extension y -1 en otro caso
   El parametro cidnumber se carga con el valor de cid number en caso de exito en
   encontrar dicho valor */
static int fix_sip_cidnumber(struct ast_channel *chan);


/* astdb_get_exten_data(const char *exten, const char *key, char *value, int len);
   obtiene de la base de datos de asterisk el valor correspondiente
   el string key asociado a la extension exten
*/
static int astdb_get_exten_data(const char *exten, const char *key, char *value, int len);


static int get_exten_name(const char *exten, char *value, int len);

/**** Virtual PBX functions ****/
inline static int find_current_vpbx(struct ast_channel *chan);
static int get_exten_vpbx(char *exten);
inline void show_vpbxs(void);


inline static int indication_comfirm(struct ast_channel *chan)
{
	int res;
	ast_safe_sleep(chan, 500);
	res = ast_playtones_start(chan, 0, indication_confirm_tone, 0);
	ast_safe_sleep(chan, 3000);
	ast_playtones_stop(chan);
	return res;
};

inline static int playback_file(struct ast_channel *chan, const char *filename)
{
	int res = 0;
	if (chan->_state != AST_STATE_UP) {
		/* Otherwise answer the line */
		res = ast_answer(chan);
	}
	if (!res) {
		ast_stopstream(chan);
		res = ast_streamfile(chan, (char *) filename, chan->language);
		if (res == 0)
			res = ast_waitstream(chan, "");
		else {
			ast_log(LOG_WARNING, "ast_streamfile failed on %s for %s\n", chan->name,
					(char *) filename);
			res = 0;
		}
		ast_stopstream(chan);
	}
	return res;
}

inline static int find_current_vpbx(struct ast_channel *chan)
{

	const char *vpbx_str = NULL;
	int vpbx = 0;

	/* Find the virtual pbx this call belongs to */
	vpbx_str = pbx_builtin_getvar_helper(chan, "VPBX");
	if (vpbx_str) {
		if (sscanf(vpbx_str, "%d", &vpbx)) {
			if (debug) {
				ast_log(LOG_NOTICE, "Call on virtual pbx %d\n", vpbx);
			}
		} else {
			ast_log(LOG_ERROR, "VPBX variable (%s) must be numeric\n", vpbx_str);
		}
	} else {
		ast_log(LOG_WARNING, "VPBX not set, using 0 as default value\n");
		vpbx = 0;				/* default value */
	}

	return vpbx;
}

static int astdb_get_exten_data(const char *exten, const char *key, char *value, int len)
{
	int res = 1;
	char family[256] = "";

	if (!value) {
		return -1;
	}
	if (ast_strlen_zero(exten)) {
		return 0;
	}
	if (ast_strlen_zero(key)) {
		return 0;
	}
	snprintf(family, sizeof(family), "%s/%s", IPC_EXTEN_TREE, exten);
	if (!ast_db_get(family, key, value, len)) {
		if (option_verbose > 2) {
			ast_verbose(VERBOSE_PREFIX_3 "AstDB value is %s for key %s at extension %s\n",
						value, key, exten);
		}
		res = 1;
	} else {
		ast_log(LOG_NOTICE, "AstDB has no value for key %s at extension %s\n", key,
				exten);
		value[0]=0;
		res = 0;
	}
	return res;
}


static int get_exten_name(const char *exten, char *buffer, int len)
{
	int res;

	if (!buffer)
		return 0;

	if (!is_number(exten)) {
		ast_log(LOG_DEBUG, "Trying to get name for non numeric extension: %s\n", exten);
		return 0;
	}
	res = astdb_get_exten_data(exten, IPC_EXTEN_INFO, buffer, len);
	if (res && !ast_strlen_zero(buffer)) {
		if (debug)
			ast_log(LOG_NOTICE, "Found name %s for %s.\n", buffer, exten);
		res = 1;
	} else {
		res = 0;
		ast_log(LOG_NOTICE, "No Name found for %s.\n", exten);
	}
	return res;
}

static int get_exten_vpbx(char *exten)
{
	char dbres[32] = "";
	int vpbx = 0;				// Default value
	int res = 0;
	res = astdb_get_exten_data(exten, IPC_EXTEN_NUMBER, dbres, sizeof(dbres));
	if (res && !ast_strlen_zero(dbres)) {
		vpbx = strtol(dbres, (char **) NULL, 10);
	} else {
		ast_log(LOG_NOTICE, "No virtual pbx for extension %s, using 0\n", exten);
		vpbx = 0;
	}
	return vpbx;
}


// Applicacion SaveClientCallInfo
#ifdef HAS_AST18
static int savecallinfo_exec(struct ast_channel *chan, const char *data)
#else
static int savecallinfo_exec(struct ast_channel *chan, void *data)
#endif
{
	MYSQL *conn = NULL;
	char *cidnumber = "", *cidname = "";
	char name_buffer[128], dnid[32];
	const char *clientid = NULL, *called_num = NULL, *dnidname = NULL, *groupid = NULL, *agent = NULL;
	char *direction;
	char sqlbuffer[IPC_QUERY_LEN];
	enum call_direction call_dir;
	int is_update;
	struct ast_module_user *u;

	u = ast_module_user_add(chan);

	memset(sqlbuffer, 0, sizeof(sqlbuffer));

	if (!ast_strlen_zero(CALLERID_NUM(chan))) {
		cidnumber = CALLERID_NUM(chan);
	}
	if (!ast_strlen_zero(CALLERID_NAME(chan))) {
		cidname = CALLERID_NAME(chan);
	}
	if (debug) {
		ast_log(LOG_NOTICE, "CallerID parsed as number='%s', name='%s'\n", cidnumber,
				cidname);
	}
	if (IS_OUTGOING_CALL(chan)) {
		call_dir = OUTGOING_CALL;
	} else if (IS_INCOMING_CALL(chan)) {
		call_dir = INCOMING_CALL;
	} else {
		call_dir = LOCAL_CALL;
	}

 	agent = pbx_builtin_getvar_helper(chan, "AGENT_NAME");
	if (agent && debug)
		ast_log(LOG_DEBUG, "Using Agent Name: %s", agent);
	groupid = pbx_builtin_getvar_helper(chan, "GROUP_ID");
	if (groupid && debug)
		ast_log(LOG_DEBUG, "Using GroupID: %s", groupid);

	switch (call_dir) {
	case INCOMING_CALL:		// Incoming call
		direction = "in";
		clientid = pbx_builtin_getvar_helper(chan, "CLIENT_DN");
		/******** Destination ID **********/
		called_num = pbx_builtin_getvar_helper(chan, "CALLEDNUMBER");
		if (!ast_strlen_zero(called_num)) {
			ast_copy_string(dnid, called_num, sizeof(dnid));
		} else if (!ast_strlen_zero(CALLERID_DNID(chan))) {
			ast_copy_string(dnid, CALLERID_DNID(chan), sizeof(dnid));
		} else {
			ast_log(LOG_DEBUG,
					"No DNID found in channel. Using current extension instead.\n");
			ast_copy_string(dnid, CHAN_REAL_EXTEN(chan), sizeof(dnid));
		};
		if (get_exten_name(CHAN_REAL_EXTEN(chan), name_buffer, sizeof(name_buffer))) {
			dnidname = ast_strdupa(name_buffer);
		} else {
			dnidname = "";
		}
		break;
	case OUTGOING_CALL:
		// Outgoing call
		direction = "out";
		clientid = pbx_builtin_getvar_helper(chan, "CLIENT_DN");
		/********** Caller ID ************/
		if (get_exten_name(cidnumber, name_buffer, sizeof(name_buffer))) {
			cidname = ast_strdupa(name_buffer);
		} else {
			cidname = "";
		}
		/******** Destination ID **********/
		ast_copy_string(dnid, CHAN_REAL_EXTEN(chan), sizeof(dnid));
		dnidname = pbx_builtin_getvar_helper(chan, "CLIENT_CN");
		if (dnidname == NULL) {
			dnidname = "";
		}
		break;
	default:
		// Local call
		direction = "lo";
		/********** Caller ID	 ************/
		if (get_exten_name(cidnumber, name_buffer, sizeof(name_buffer))) {
			cidname = ast_strdupa(name_buffer);
			if (debug)
				ast_log(LOG_NOTICE, "Found name %s for %s.\n", cidname, cidnumber);
		} else {
			cidname = "";
		}
		/******** Destination ID **********/
		ast_copy_string(dnid, CHAN_REAL_EXTEN(chan), sizeof(dnid));
		dnidname = "";
		if (get_exten_name(dnid, name_buffer, sizeof(name_buffer))) {
			dnidname = ast_strdupa(name_buffer);
			if (debug)
				ast_log(LOG_DEBUG, "Found name %s for %s.\n", dnidname, dnid);
		}
		ast_copy_string(dnid, CHAN_REAL_EXTEN(chan), sizeof(dnid));
	}
	is_update = ast_true(pbx_builtin_getvar_helper(chan, "SAVED_CALL"));
	if (callinfo.usemanager) {
		if (is_update) {
			manager_event(EVENT_FLAG_AGENT, "UpdateClientCall",
						  "Direction: %s\r\n"
						  "Uniqueid: %s\r\n"
						  "CallerID: %s\r\n"
						  "CallerIDName: %s\r\n"
						  "DNID: %s\r\n"
						  "DNIDName: %s\r\n"
						  "AgentName: %s\r\n"
						  "GroupID: %s\r\n",
							direction, chan->uniqueid, 
							cidnumber, cidname, dnid, dnidname, agent ? agent : "", groupid ? groupid : "");
		} else {
			manager_event(EVENT_FLAG_AGENT, "SaveClientCall",
						  "Direction: %s\r\n"
						  "Uniqueid: %s\r\n"
						  "CallerID: %s\r\n"
						  "CallerIDName: %s\r\n"
						  "DNID: %s\r\n"
						  "DNIDName: %s\r\n"
						  "ClientID: %s\r\n",
							direction, chan->uniqueid,
						  cidnumber, cidname, dnid, dnidname, clientid ? clientid : "");
		}
	}
	if (callinfo.usedb) {
	/***** BEGIN MUTEX BLOCK *******/
		ipc_mutex_lock();
		conn = ipc_sql_reconnect();
		if (conn) {
			char *end = sqlbuffer;
			if (is_update) {
				end = stpcpy(sqlbuffer, "UPDATE callinfo SET cid=");
				*end++ = '\'';
				end += mysql_real_escape_string(conn, end, cidnumber, strlen(cidnumber));
				*end++ = '\'';
				end = stpcpy(end, ",cidname=");
				*end++ = '\'';
				end += mysql_real_escape_string(conn, end, cidname, strlen(cidname));
				*end++ = '\'';
				end = stpcpy(end, ",dnid=");
				*end++ = '\'';
				end += mysql_real_escape_string(conn, end, dnid, strlen(dnid));
				*end++ = '\'';
				end = stpcpy(end, ",dnidname=");
				*end++ = '\'';
				end += mysql_real_escape_string(conn, end, dnidname, strlen(dnidname));
				*end++ = '\'';
				end = stpcpy(end, " WHERE uniqueid=");
				*end++ = '\'';
				end +=
					mysql_real_escape_string(conn, end, chan->uniqueid,
											 strlen(chan->uniqueid));
				*end++ = '\'';
			} else {
				end = stpcpy(sqlbuffer,
							 "INSERT INTO callinfo (uniqueid,direction,cid,cidname,dnid,dnidname,clientid) values (");
				*end++ = '\'';
				end += mysql_real_escape_string(conn, end, chan->uniqueid, strlen(chan->uniqueid));
				*end++ = '\'';
				*end++ = ',';
				*end++ = '\'';
				end += mysql_real_escape_string(conn, end, direction, strlen(direction));
				*end++ = '\'';
				*end++ = ',';
				*end++ = '\'';
				end += mysql_real_escape_string(conn, end, cidnumber, strlen(cidnumber));
				*end++ = '\'';
				*end++ = ',';
				*end++ = '\'';
				end += mysql_real_escape_string(conn, end, cidname, strlen(cidname));
				*end++ = '\'';
				*end++ = ',';
				*end++ = '\'';
				end += mysql_real_escape_string(conn, end, dnid, strlen(dnid));
				*end++ = '\'';
				*end++ = ',';
				*end++ = '\'';
				end += mysql_real_escape_string(conn, end, dnidname, strlen(dnidname));
				*end++ = '\'';
				*end++ = ',';
				*end++ = '\'';
				if (clientid)
					end +=
						mysql_real_escape_string(conn, end, clientid, strlen(clientid));
				*end++ = '\'';
				*end++ = ')';
			}
			*end = '\0';
			if (mysql_real_query(conn, sqlbuffer, (size_t)(end - sqlbuffer))) {
				ast_log(LOG_ERROR, "Database query failed: %s.\n", sqlbuffer);
			}
		} else {
			ast_log(LOG_ERROR, "Unable to connect to MySQL server\n");
		};
		/***** END MUTEX BLOCK *******/
		ipc_mutex_unlock();
	}
	pbx_builtin_setvar_helper(chan, "SAVED_CALL", "1");
	ast_module_user_remove(u);
	return 0;
}


static int fix_sip_cidnumber(struct ast_channel *chan)
{
	MYSQL_RES *sqlres = NULL;
	MYSQL_ROW row;
	char sql[IPC_QUERY_LEN] = "";
	int res = 0;

	if (!CALLERID_NUM(chan)) {	/* No cid number found in channel */
		ast_log(LOG_NOTICE, "No Caller ID info in channel %s\n", chan->name);
		return -1;
	}

	snprintf(sql, (int) sizeof(sql),
			 "SELECT number FROM extension WHERE protocol LIKE 'SIP' AND channel LIKE '%s@%%'",
			 CALLERID_NUM(chan));

	sqlres = ipc_sql_exec_query(sql);

	if ((sqlres != NULL) && (mysql_num_rows(sqlres) > 0)) {
		row = mysql_fetch_row(sqlres);
		if ((row != NULL) && !ast_strlen_zero(row[0])) {
			if (debug) {
				ast_log(LOG_NOTICE, "Found extension = %s for user %s\n", row[0],
						CALLERID_NUM(chan));
			};
			ast_set_callerid(chan, row[0],
							 ast_strlen_zero(CALLERID_NAME(chan)) ? NO_CID_NAME : NULL,
							 NULL);
		} else {
			res = -1;
		};
		mysql_free_result(sqlres);
	} else if (debug) {
		ast_log(LOG_NOTICE, "No extension found for SIP user %s\n", CALLERID_NUM(chan));
		res = -1;
	};
	return res;
}

#ifdef HAS_AST18
static int setcidext_exec(struct ast_channel *chan, const char *data)
#else
static int setcidext_exec(struct ast_channel *chan, void *data)
#endif
{
	if (!strcmp(ast_channel_type(chan), "SIP")) {
		fix_sip_cidnumber(chan);
	};
	return 0;
}

#ifdef HAS_AST18
static int callforward_exec(struct ast_channel *chan, const char*data)
#else
static int callforward_exec(struct ast_channel *chan, void *data)
#endif
{
	int res = 0;
	int enable = 0;
	char newexten[32];
	char family[IPC_MAX_LEN];
	const char *fwd_type = NULL;
	int usage = 0;
	char *parse = NULL;
	struct ast_module_user *u;

	AST_DECLARE_APP_ARGS(args, AST_APP_ARG(type);
						 AST_APP_ARG(mode);
		);

	if (ast_strlen_zero(data)) {
		usage = 1;
	}
	parse = ast_strdupa(data);
	if (!parse) {
		ast_log(LOG_ERROR, "Unable to dupe data!");
		return -1;
	}

	AST_STANDARD_APP_ARGS(args, parse);

	if (ast_strlen_zero(args.type)) {
		usage = 1;
	} else {
		if (strncasecmp(args.type, "a", 1))
			fwd_type = CALL_FORWARD_ALL;
		else if (strncasecmp(args.type, "b", 1))
			fwd_type = CALL_FORWARD_BUSY;
		else if (strncasecmp(args.type, "u", 1))
			fwd_type = CALL_FORWARD_UNAVAIL;
		else
			usage = 1;
	};

	if (usage) {
		ast_log(LOG_ERROR,
				"Call Forward requires a parameter: type[" AST_STANDARD_APP_DELIM "<on|off>]. type: (a|u|b).\n");
		return -1;
	}

	u = ast_module_user_add(chan);

	if (ast_strlen_zero(args.mode))
		enable = 1;
	else
		enable = ast_true(args.mode);
	memset(newexten, 0, sizeof(newexten));

	if (!strcmp(ast_channel_type(chan), "SIP"))
		res = fix_sip_cidnumber(chan);

	if (ast_strlen_zero(CALLERID_NUM(chan))) {
		ast_log(LOG_ERROR, "Found empty CID number, unable to exec.\n");
		res = -1;
	};
	/* Answer if need be */
	if (chan->_state != AST_STATE_UP) {
		res = ast_answer(chan);
	}
	snprintf(family, sizeof(family), "%s/%s", CALL_FORWARD_BASE, fwd_type);
	if (!res) {
		/* Stop anything playing */
		ast_stopstream(chan);

		if (enable) {			/* Enable call forwarding in this extension */
			ast_safe_sleep(chan, 500);
			if ((res = ast_streamfile(chan, "beep", chan->language))) {
				ast_log(LOG_WARNING, "Unable to stream beep file.\n");
			} else {
				res = ast_waitstream(chan, "");
			}
			ast_stopstream(chan);

			res =
				ast_readstring(chan, newexten, sizeof(newexten) - 1, digit_timeout,
							   first_digit_timeout, "#");

			if ((res < 0) ||
				!ast_exists_extension(chan, chan->context, newexten, 1,
									  CALLERID_NUM(chan))) {
				/* Error or received no valid extension */
				if (debug) {
					if (!ast_strlen_zero(newexten))
						ast_log(LOG_WARNING, "Invalid extension: %s.\n", newexten);
					else
						ast_log(LOG_WARNING, "No digits received!!");
				}
				res = ast_streamfile(chan, "pbx-invalid", chan->language);
				if (!res) {
					res = ast_waitstream(chan, "");
				}
				ast_stopstream(chan);
				goto end;
			} else {			/* We got a valid extension */
				if (ast_db_put(family, CALLERID_NUM(chan), newexten)) {
					ast_log(LOG_WARNING,
							"Error writing value to database entry: %s/%s.\n", family,
							CALLERID_NUM(chan));
					res = -1;
				} else {
					ast_say_number(chan, atoi((char *) newexten), "", chan->language,
								   NULL);
					res = indication_comfirm(chan);
				}
			};
		} else {				/* Disable call forwarding in this extension */
			if (ast_db_del(family, CALLERID_NUM(chan))) {
				ast_log(LOG_WARNING, "Error deleting database entry: %s/%s.\n",
						family, CALLERID_NUM(chan));
				res = -1;
			} else {
				res = indication_comfirm(chan);
			};
		};
	} else {
		ast_log(LOG_WARNING, "Unable to answer channel %s.\n", chan->name);
	};

  end:
	ast_module_user_remove(u);
	return res;
}

#ifdef HAS_AST18
static int get_pin_profile_exec(struct ast_channel *chan, const char *data)
#else
static int get_pin_profile_exec(struct ast_channel *chan, void *data)
#endif
{
	int res = 0;
	MYSQL_RES *sqlres = NULL;
	MYSQL_ROW rowval;
	MYSQL_FIELD *fields = NULL;
	char query[IPC_QUERY_LEN];
	int numfields;
	char pin[10];
	struct ast_module_user *u;

	if (ast_strlen_zero(data) || atoi(data) < 0) {
		ast_log(LOG_WARNING, "GetPinProfile requires a numeric argument: pin\n");
		return -1;
	}
	ast_copy_string(pin, data, sizeof(pin));

	u = ast_module_user_add(chan);

	/* Busco los permisos asociados a este pin */
	if (schemaversion == 1) {
		snprintf(query, sizeof(query),
						 "SELECT username,local,longdist,cellular,inter FROM authcalls WHERE pin='%s' AND enabled", pin);
	}	else {
		snprintf(query, sizeof(query),
						"SELECT username,local,longdist,cellular,inter FROM authcalls a, users u WHERE pin='%s' AND enabled AND a.user_id=u.user_id", pin); 
	}

	sqlres = ipc_sql_exec_query(query);
	if (sqlres && (mysql_num_rows(sqlres) > 0)) {
		/* Hubo resultado en la consulta */
		numfields = mysql_num_fields(sqlres);
		fields = mysql_fetch_fields(sqlres);
		/* Obtengo la primer fila */
		rowval = mysql_fetch_row(sqlres);
		int i;
		for (i = 0; i < numfields; i++) {
			if (rowval[i]) {
				if (!strcmp(fields[i].name, "username")) {
					// username is not defined
					if (ast_strlen_zero(rowval[i])) {
						pbx_builtin_setvar_helper(chan, "AUTH_user", "");
						break;
					} else {
						pbx_builtin_setvar_helper(chan, "AUTH_user", rowval[i]);
					}
				} else if (!strcmp(fields[i].name, "local")) {
					pbx_builtin_setvar_helper(chan, "AUTH_local", rowval[i]);
				} else if (!strcmp(fields[i].name, "longdist")) {
					pbx_builtin_setvar_helper(chan, "AUTH_longdist", rowval[i]);
				} else if (!strcmp(fields[i].name, "cellular")) {
					pbx_builtin_setvar_helper(chan, "AUTH_cellular", rowval[i]);
				} else if (!strcmp(fields[i].name, "inter")) {
					pbx_builtin_setvar_helper(chan, "AUTH_inter", rowval[i]);
				}
			}
		}
	} else {
		pbx_builtin_setvar_helper(chan, "AUTH_user", "");
	}
	ast_module_user_remove(u);
	return res;
}

#ifdef HAS_AST18
static int get_exten_profile_exec(struct ast_channel *chan, const char *data)
#else
static int get_exten_profile_exec(struct ast_channel *chan, void *data)
#endif
{
	int res = 0;
	MYSQL_RES *sqlres = NULL;
	MYSQL_ROW rowval;
	MYSQL_FIELD *fields = NULL;
	char query[IPC_QUERY_LEN];
	int numfields;
	int pin = 0;
	struct ast_module_user *u;

	if (ast_strlen_zero(data) || (pin = atoi(data)) < 0) {
		ast_log(LOG_WARNING, "GetExtenProfile requires an extension argument\n");
		return -1;
	}

	u = ast_module_user_add(chan);

	/* Busco los permisos asociados a este pin */
	snprintf(query, sizeof(query),
			 "SELECT local,longdist,cellular,inter FROM extension WHERE number='%s'",
			 (char*)data);
	sqlres = ipc_sql_exec_query(query);
	if (sqlres && (mysql_num_rows(sqlres) > 0)) {
		/* Hubo resultado en la consulta */
		numfields = mysql_num_fields(sqlres);
		fields = mysql_fetch_fields(sqlres);
		/* Obtengo la primer fila */
		rowval = mysql_fetch_row(sqlres);
		int i;
		for (i = 0; i < numfields; i++) {
			if (rowval[i]) {
				if (!strcmp(fields[i].name, "local")) {
					pbx_builtin_setvar_helper(chan, "AUTH_local", rowval[i]);
				} else if (!strcmp(fields[i].name, "longdist")) {
					pbx_builtin_setvar_helper(chan, "AUTH_longdist", rowval[i]);
				} else if (!strcmp(fields[i].name, "cellular")) {
					pbx_builtin_setvar_helper(chan, "AUTH_cellular", rowval[i]);
				} else if (!strcmp(fields[i].name, "inter")) {
					pbx_builtin_setvar_helper(chan, "AUTH_inter", rowval[i]);
				}
			}
		}
	}
	ast_module_user_remove(u);
	return res;
}

#ifdef HAS_AST18
static int dialhunt_exec(struct ast_channel *chan, const char *data)
#else
static int dialhunt_exec(struct ast_channel *chan, void *data)
#endif
{
	int res = 0;
	MYSQL_RES *sqlres = NULL;
	MYSQL_ROW rowval;
	MYSQL_FIELD *fields = NULL;
	int numfields;
	char dialargs[IPC_MAX_LEN];
	char query[IPC_QUERYLONG_LEN];
	char var_buffer[32];
	int vpbx = 0;
	int huntgroup = 0;
	long unsigned numlines = 0;
	struct line_info *line;
	int ready = 0;
	const char *dialstatus = "INVALIDARGS";
	int failonbusy;
	struct ast_frame *frm;
	time_t rstart;
	struct ast_module_user *u;
	char *parse = NULL;
	struct ast_flags flags = { 0, };
	char last_id[4];;
	char family[256] = "";
	char line_field[16] = "pstnline_id";
	char hunt_field[16] = "pstnline_id";
	  
	AST_DECLARE_APP_ARGS(args,
						 AST_APP_ARG(huntgroup);
						 AST_APP_ARG(number);
						 AST_APP_ARG(opts);
						 AST_APP_ARG(dialopts);
		);

	if (ast_strlen_zero(data)) {
		ast_log(LOG_WARNING,
				"DialHuntGroup requires an argument (hunt" AST_STANDARD_APP_DELIM "number" AST_STANDARD_APP_DELIM "options" AST_STANDARD_APP_DELIM "dialopts)\n");
		pbx_builtin_setvar_helper(chan, "DIALSTATUS", dialstatus);
		return -1;
	}

	parse = ast_strdupa(data);
	if (!parse) {
		ast_log(LOG_ERROR, "Unable to dupe data :(\n");
		return -1;
	}
	AST_STANDARD_APP_ARGS(args, parse);

	if (ast_strlen_zero(args.huntgroup) || ((huntgroup = atoi(args.huntgroup)) < 0)) {
		ast_log(LOG_ERROR, "DialHuntGroup requires a numeric huntgroup\n");
		pbx_builtin_setvar_helper(chan, "DIALSTATUS", dialstatus);
		return -1;
	}

	if (ast_strlen_zero(args.number)) {
		ast_log(LOG_ERROR, "DialHuntGroup requires a number to dial\n");
		pbx_builtin_setvar_helper(chan, "DIALSTATUS", dialstatus);
		return -1;
	}
	// Everything ok, let's increment module usage
	u = ast_module_user_add(chan);

	line = malloc(sizeof(struct line_info));
	memset(var_buffer, 0, sizeof(var_buffer));

	if (!ast_strlen_zero(args.opts)) {
		ast_app_parse_options(dialhunt_exec_opts, &flags, NULL, args.opts);
	}

 	if (!ast_test_flag(&flags, HUNT_FAILONBUSY) && !ast_test_flag(&flags, HUNT_NOFAILONBUSY)) {
		failonbusy = dialhunt.failonbusy;
		if (debug)
			ast_log(LOG_NOTICE,	"DialHuntGroup called without failonbusy option, defaults to config file value: failonbusy=%d.\n", failonbusy);
	} else if (ast_test_flag(&flags, HUNT_FAILONBUSY) && ast_test_flag(&flags, HUNT_NOFAILONBUSY)) {
		ast_log(LOG_ERROR, "You cannot mix option '1' whith '0|c'. Aborted\n");
		goto out;
	} else {
		failonbusy = ast_test_flag(&flags, HUNT_FAILONBUSY) || !ast_test_flag(&flags, HUNT_NOFAILONBUSY);
		if(debug)
			ast_log(LOG_NOTICE, "Fail On Busy: %d.\n", failonbusy);
	}

	if (ast_strlen_zero(args.dialopts)) {
		if (ast_strlen_zero(dialhunt.dialopts)) {
			args.dialopts = "";
			if (debug)
				ast_log(LOG_NOTICE, "Dialing without dial options.\n");
		} else {
			args.dialopts = dialhunt.dialopts;
			if (debug)
				ast_log(LOG_NOTICE,
						"No dial options. Taking defaults from app config.\n");
		}
	}

	if (ast_strlen_zero(CALLERID_NUM(chan))) {
		ast_log(LOG_WARNING,
				"Unable to get valid extension from cid, using virtual pbx 0");
	} else {
		vpbx = get_exten_vpbx(CALLERID_NUM(chan));
	}

	snprintf(var_buffer, sizeof(var_buffer), "%d", vpbx);
	pbx_builtin_setvar_helper(chan, "VPBX", var_buffer);

	/* Compatibilidad con DialHuntGroup < 1.4.6 */
	if (schemaversion == 1) {
	  snprintf(line_field, sizeof(line_field), "number");
	  snprintf(hunt_field, sizeof(hunt_field), "line_number");
	}
	/* Busco las lineas en este huntgroup */
	if (ast_test_flag(&flags, HUNT_DIAL_ROTASC) || ast_test_flag(&flags, HUNT_DIAL_ROTDESC)) {
		snprintf(family, sizeof(family), "%s/%u", IPC_HUNT_TREE, huntgroup);
		if (ast_db_get(family, "last_id", last_id, sizeof(last_id)))
			ast_copy_string(last_id, "0", 2);
		snprintf(query, sizeof(query),
				 "(SELECT 1 as sort,id,number,protocol,line_prefix,line_postfix FROM hunt_line h, pstnline l WHERE l.%s=h.%s AND hunt_number=%u AND enabled=1 AND busy=0 AND id%s%s) UNION (SELECT 2,id,number,protocol,line_prefix,line_postfix FROM hunt_line h, pstnline l WHERE l.%s=h.%s AND hunt_number=%u AND enabled=1 AND busy=0 AND id%s%s ) ORDER by sort,id %s",
				 line_field, hunt_field,
				 huntgroup,
				 ast_test_flag(&flags, HUNT_DIAL_ROTASC) ? ">" : "<",
				 last_id, 
				 line_field, hunt_field,
				 huntgroup, 
				 ast_test_flag(&flags, HUNT_DIAL_ROTASC) ? "<=" : ">=",
				 last_id,
				 ast_test_flag(&flags, HUNT_DIAL_ROTDESC) ? "DESC" : "" );
	} else {
		snprintf(query, sizeof(query),
				 "SELECT number,protocol,line_prefix,line_postfix FROM pstnline as l, hunt_line as h WHERE l.%s = h.%s AND h.hunt_number = %d  AND enabled = 1 AND busy = 0 ORDER BY id %s",
				 line_field, hunt_field,
				 huntgroup, 
				 ast_test_flag(&flags, HUNT_DIAL_SEQDESC) ? "DESC" : "" );
	}

	sqlres = ipc_sql_exec_query(query);
	if (sqlres) {
		numlines = mysql_num_rows(sqlres);
	} else {
		res = -1;
	}

	if ((res == 0) && (numlines > 0)) {
		/* Hubo resultado en la consulta */
		/* y hay lineas libres en este huntgroup */
		snprintf(var_buffer, sizeof(var_buffer), "%d", huntgroup);
		pbx_builtin_setvar_helper(chan, "HUNTGROUP", var_buffer);

		if (debug) {
			ast_log(LOG_NOTICE, "Have %lu free lines on huntgroup %d \n", numlines, huntgroup);
		}
		// Itero sobre las lineas encontradas en este HUNTGROUP
		int i;
		while ((rowval = mysql_fetch_row(sqlres))) {
			memset(line, 0, sizeof(struct line_info));
			numfields = mysql_num_fields(sqlres);
			fields = mysql_fetch_fields(sqlres);
			for (i = 0; i < numfields; i++) {
				if (rowval[i]) {
				  if (!strcmp(fields[i].name, "number")) {
						ast_copy_string(line->number, rowval[i], sizeof(line->number));
					} else if (!strcmp(fields[i].name, "protocol")) {
						ast_copy_string(line->protocol, rowval[i], sizeof(line->protocol));
					} else if (!strcmp(fields[i].name, "line_prefix")) {
						ast_copy_string(line->prefix, rowval[i], sizeof(line->prefix));
					} else if (!strcmp(fields[i].name, "line_postfix")) {
						ast_copy_string(line->postfix, rowval[i], sizeof(line->postfix));
					} else if (!strcmp(fields[i].name, "id")) {
						ast_copy_string(last_id, rowval[i], sizeof(last_id));
						snprintf(family, sizeof(family), "%s/%u", IPC_HUNT_TREE, huntgroup);
						ast_db_put(family, "last_id", last_id);
					}
				}
			}
			memset(dialargs, 0, sizeof(dialargs));
			if (!build_dial_string(line, args.number, dialargs, sizeof(dialargs))) {
				char *cursor = dialargs + strlen(dialargs);
				size_t left = sizeof(dialargs) - strlen(dialargs);
				if (!ast_strlen_zero(args.dialopts) && (left > 2)) {
					ast_build_string(&cursor, &left, AST_STANDARD_APP_DELIM);
					ast_build_string(&cursor, &left, args.dialopts);
				}
				if (debug)
					ast_log(LOG_NOTICE, "Dialing: %s\n", dialargs);
				res = pbx_exec(chan, appDial, dialargs);

				if (res == 0) {	//Dial application returned without hangup
					dialstatus = pbx_builtin_getvar_helper(chan, "DIALSTATUS");
					if (debug) {
						ast_log(LOG_NOTICE, "Dial returned with status: %s\n",
								dialstatus);
					}
					if (!strcmp(dialstatus, "BUSY")) {	// Dial returns BUSY on this line
						if (failonbusy) {
							ast_indicate(chan, AST_CONTROL_BUSY);
							/* give em busy tone for several seconds */
							time(&rstart);
							while (time(NULL) < rstart + 10) {
								if (ast_waitfor(chan, -1) < 0)
									break;
								frm = ast_read(chan);
								if (!frm)
									break;
								ast_frfree(frm);
							}
							ast_playtones_stop(chan);
							goto out;
						} else {
							ast_log(LOG_NOTICE,
									"Dial returned BUSY on line %s. Going to next line...\n",
									line->number);
						}
					} else if (!strcmp(dialstatus, "NOANSWER")) {	// Dial Not Answered
						ready = 1;
						break;
					}
				} else {
					ast_log(LOG_DEBUG, "Dial result: %d\n", res);
					ready = 1;
					break;
				}
			}
		}						// Fin iteracion sobre HUNTGROUP
	}							// Fin if(res == 0 && numlines > 0)
	if (!ready) {				// Dial not completed but we have no more lines
		ast_log(LOG_NOTICE, "No more lines... DialHuntGroup returned NOMORELINES\n");
		pbx_builtin_setvar_helper(chan, "DIALHUNTGROUPSTATUS", "NOMORELINES");
	}
  out:
	if (sqlres) {
		mysql_free_result(sqlres);
		sqlres = NULL;
	}
	if (line)
		free(line);
	ast_module_user_remove(u);
	return res;
}

#ifdef HAS_AST18
static int vpbxgoto_exec(struct ast_channel *chan, const char *data)
#else
static int vpbxgoto_exec(struct ast_channel *chan, void *data)
#endif
{
	int res = 0;
	struct virtual_pbx *p = NULL;
	char target[256];
	int vpbx = 0;
	char destflag;
	struct ast_module_user *u;

	if (ast_strlen_zero(data)) {
		ast_log(LOG_WARNING, "VirtualPBXGoto requires a flag argument)\n");
		return -1;
	}

	u = ast_module_user_add(chan);

	vpbx = find_current_vpbx(chan);
	p = get_virtual_pbx(vpbx);
	if (!p) {
		ast_log(LOG_ERROR, "Virtual PBX %d not configured!", vpbx);
		res = -1;
		goto out;
	}

	destflag = ((char *) data)[0];
	switch (destflag) {
	case 'm':
		ast_verbose(VERBOSE_PREFIX_3 "Virtual PBX Goto:  mainmenu\n");
		res = ast_parseable_goto(chan, (char *) p->mainmenu);
		break;
	case 'o':
		ast_verbose(VERBOSE_PREFIX_3 "Virtual PBX Goto:  operator\n");
		res = ast_parseable_goto(chan, (char *) p->operator);
		break;
	case 'v':
		ast_verbose(VERBOSE_PREFIX_3 "Virtual PBX Goto:  voicemail\n");
		if (IS_INCOMING_CALL(chan))
			snprintf(target, sizeof(target), "%s" AST_STANDARD_APP_DELIM "%s", (char *) p->vmsincoming, "s" AST_STANDARD_APP_DELIM "1");
		else
			snprintf(target, sizeof(target), "%s" AST_STANDARD_APP_DELIM "%s", (char *) p->vmslocal, "s" AST_STANDARD_APP_DELIM "1");
		res = ast_parseable_goto(chan, target);
		break;
	default:
		ast_log(LOG_WARNING, "VirtualPBXGoto argument must be 'm', 'o' or 'v')\n");
		res = -1;
		break;
	}
  out:
	ast_module_user_remove(u);
	return res;
}

inline int build_dial_string(struct line_info *line, const char *number, char *dialstr,const int len)
{
	int ret = 0;
	if (!strcasecmp(line->protocol, "zap") || !strcasecmp(line->protocol, "dahdi")) {
		snprintf(dialstr, len, "%s/%s/%s", line->protocol, line->prefix, number);
	} else if (!strcasecmp(line->protocol, "sip")) {
		snprintf(dialstr, len, "sip/%s%s%s", line->prefix, number, line->postfix);
	} else if (!strcasecmp(line->protocol, "iax2")) {
		snprintf(dialstr, len, "%s/%s/%s", line->protocol, line->prefix, number);
	} else if (!strcasecmp(line->protocol, "h323")) {
		snprintf(dialstr, len, "h323/%s%s", line->prefix, number);
	} else if (!strcasecmp(line->protocol, "oh323")) {
		snprintf(dialstr, len, "oh323/%s%s", line->prefix, number);
	} else if (!strcasecmp(line->protocol, "capi")) {
		if (!ast_strlen_zero(line->postfix)) {
			snprintf(dialstr, len, "capi/%s/%s/%s", line->prefix, number, line->postfix);
		} else {
			snprintf(dialstr, len, "capi/%s/%s", line->prefix, number);
		}
	} else if (!strcasecmp(line->protocol, "vpb")) {
		snprintf(dialstr, len, "vpb/%s/%s", line->prefix, number);
	} else {
		ast_log(LOG_WARNING, "Unknown protocol %s on line %s\n", line->protocol,
				line->number);
		ret = -1;
	}
	if (!ret && debug) {
		ast_log(LOG_NOTICE, "Line: protocol => %s prefix => %s postfix => %s\n",
				line->protocol, line->prefix, line->postfix);
		ast_log(LOG_NOTICE, "Channel Dial String: %s\n", dialstr);
	}
	return ret;
}


/* load_config:
   Load configuration from dialipcontact.conf 
   Returns 0 on success or -1 when an error is found.\n"
*/
static int load_config(void)
{
	struct ast_config *config;
	struct ast_variable *var;
	char config_path[PATH_MAX] = "";

	snprintf(config_path, sizeof(config_path), "%s/%s", IPCONTACT_CONFDIR, config_file);
	config_path[sizeof(config_path) - 1] = 0;

  /***** BEGIN MUTEX BLOCK **********/
	ipc_mutex_lock();
#if defined(HAS_AST16) || defined(HAS_AST18)
	struct ast_flags config_flags = { CONFIG_FLAG_WITHCOMMENTS };
	config = ast_config_load((char *) config_path, config_flags);
#else
	config = ast_config_load((char *) config_path);
#endif
	if (!config) {
		ast_log(LOG_WARNING, "Unable to load config file: %s. Using default values.\n", config_path);
		goto end;
	};

	if (ast_true(ast_variable_retrieve(config, "general", "debug"))) {
		ast_log(LOG_NOTICE, "%s loaded in debug mode\n", MODULE_DESC);
		debug = 1;
	} else {
		debug = 0;
	}

	if (debug)
		ast_log(LOG_NOTICE, "Using schema version %d\n", schemaversion);

	// Parsing callinfo section
	var = ast_variable_browse(config, "callinfo");
	while (var) {
		if (!strcasecmp(var->name, "usedb")) {
			callinfo.usedb = ast_true(var->value) ? 1:0;
		}
		if (!strcasecmp(var->name, "usemanager")) {
			callinfo.usemanager = ast_true(var->value) ? 1:0;
		}
		var = var->next;
	}

	// Parsing dialhunt section
	var = ast_variable_browse(config, "dialhuntgroup");
	while (var) {
		if (!strcasecmp(var->name, "failonbusy")) {
			dialhunt.failonbusy = ast_true(var->value) ? 1:0;
		}
		if (!strcasecmp(var->name, "dialopts")) {
			ast_copy_string(dialhunt.dialopts, var->value, sizeof(dialhunt.dialopts));
		}
		var = var->next;
	}
  end:
	ast_config_destroy(config);
	ipc_mutex_unlock();
  /***** END MUTEX BLOCK **********/

	return 0;
}

/*
 * Standard module functions ...
 */
static int load_module(void)
{
	int res = 0;
	char *ipcontact_loaded = ast_module_helper("", "res_ipcontact", 0, 0, 0, 0);
	free(ipcontact_loaded);

	if (!ipcontact_loaded) {
		ast_log(LOG_ERROR, "app_dialipcontact.so depends upon res_ipcontact.so\n");
		return AST_MODULE_LOAD_DECLINE;
	}
	if (!(appDial = pbx_findapp("Dial"))) {
		return AST_MODULE_LOAD_DECLINE;
	} else {
		if (load_config())
			return AST_MODULE_LOAD_DECLINE;
	}
#ifdef HAS_XMLDOC
	res |= ast_register_application_xml(app_dialhunt, dialhunt_exec);
	res |= ast_register_application_xml(app_savecallinfo, savecallinfo_exec);
	res |= ast_register_application_xml(app_vpbxgoto, vpbxgoto_exec);
	res |= ast_register_application_xml(app_get_pin_profile, get_pin_profile_exec);
	res |= ast_register_application_xml(app_get_exten_profile, get_exten_profile_exec);
	res |= ast_register_application_xml(app_callforward, callforward_exec);
	res |= ast_register_application_xml(app_setcidext, setcidext_exec);
#else
	res |= ast_register_application(app_dialhunt, dialhunt_exec, synopsis_dialhunt, descrip_dialhunt);
	res |= ast_register_application(app_savecallinfo, savecallinfo_exec, synopsis_savecallinfo, descrip_savecallinfo);
	res |= ast_register_application(app_vpbxgoto, vpbxgoto_exec, synopsis_vpbxgoto, descrip_vpbxgoto);
	res |= ast_register_application(app_get_pin_profile, get_pin_profile_exec, synopsis_get_pin_profile, descrip_get_pin_profile);
	res |= ast_register_application(app_get_exten_profile, get_exten_profile_exec, synopsis_get_exten_profile, descrip_get_exten_profile);
	// Estas 2 ultimas aplicaciones no se usan, eliminarlas algun dia
	res |= ast_register_application(app_callforward, callforward_exec, synopsis_callforward, descrip_callforward);
	res |= ast_register_application(app_setcidext, setcidext_exec, synopsis_setcidext, descrip_setcidext);
#endif
	return AST_MODULE_LOAD_SUCCESS;
}

static int unload_module(void)
{
	int res = 0;
	res |= ast_unregister_application(app_dialhunt);
	res |= ast_unregister_application(app_setcidext);
	res |= ast_unregister_application(app_savecallinfo);
	res |= ast_unregister_application(app_vpbxgoto);
	res |= ast_unregister_application(app_callforward);
	res |= ast_unregister_application(app_get_pin_profile);
	res |= ast_unregister_application(app_get_exten_profile);
	return res;
};


AST_MODULE_INFO_STANDARD(ASTERISK_GPL_KEY, MODULE_DESC);
