/*
 * Asterisk -- A telephony toolkit for Linux.
 * Copyright (C) 1999-2005, Digium, Inc.
 * Mark Spencer <markster@digium.com>  - Asterisk main author
 *
 * res_ipcontact.c <Plugin for IPContact System configuration>
 * Copyright (c) 2004-2008, IPContact Software SRL.  All rights reserved.
 *
 * Luis Vazquez <luis@ipcontact.com.uy>
 * Daniel Ferrer <daniel@ipcontact.com.uy>
 *
 * $Id: res_ipcontact.c,v 1.11 2007-07-10 00:16:50 ipcontact Exp $
 *
 * This code is released by the author with no restrictions on usage.
 *
 */

/*! \file
 *
 * \brief IPContact System Resources
 *
 * \author Luis Vazquez <luis@ipcontact.com.uy>
 * \author Daniel Ferrer <daniel@ipcontact.com.uy>
 * 
 * \ingroup resources
 */

/*** MODULEINFO
     <depend>mysqlclient</depend>
***/


#include <asterisk.h>

#include <asterisk/channel.h>
#include <asterisk/logger.h>
#include <asterisk/config.h>
#include <asterisk/module.h>
#include <asterisk/lock.h>
#include <asterisk/options.h>
#include <asterisk/cli.h>
#include <asterisk/utils.h>
#include <asterisk/pbx.h>
#include <asterisk/app.h>
#include <asterisk/astdb.h>

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <ctype.h>

#include <mysql/mysql_version.h>
#include <mysql/errmsg.h>


#include "res_ipcontact.h"

/*** DOCUMENTATION

<function name="NORM_ANTEL" language="en_US">
		<synopsis>
			Normalize ANTEL callerid numbers
		</synopsis>
		<syntax>
			<parameter name="number" required="true">
				<para>Phone number to normalize</para>
			</parameter>
		</syntax>
		<description>
			 <para>Normalize a callerid number as received from ANTEL. Returned number must follow the pattern:</para>
			 <para>[local code]XXXX for national calls</para>
			 <para>09XXXXXXX for cellular numbers</para>
			 <para>1XX. , 0800XXXX and 0900XXXX are not changed</para>
		</description>
	</function>

	<function name="PROFILE_AUTH" language="en_US">
		<synopsis>
			Check profile auth for a given call type, from current extension
		</synopsis>
		<syntax>
			<parameter name="type" required="true">
				<para>Call Type:</para>
				<enumlist>
					 <enum name="local">
						 <para>Local calls (default)</para>
					 </enum>
					 <enum name="long">
						 <para>Long distance calls</para>
					 </enum>
					 <enum name="cell">
						 <para>Cellular calls</para>
					 </enum>
					 <enum name="inter">
						 <para>International calls</para>
					 </enum>
			    </enumlist>
			</parameter>
		</syntax>
		<description>
			 <para>Check if the curren extension has permissions to initiate a call of the given type.</para>
			 <para>It checks the permission looking on IPContact extension tree on astdb (ipcontact/extension/<replaceable>extension</replaceable>/profile)</para>
			 <para>Valid call types values are:</para>
			 <enumlist>
				<enum name="local"/>
				<enum name="longdist"/>
				<enum name="cellular"/>
				<enum name="inter"/>
			</enumlist>
		</description>
	</function>

	<function name="EXTEN_PIN" language="en_US">
		<synopsis>
			Gets or sets extension's authentication pin
		</synopsis>
		<syntax>
			<parameter name="exten" required="true">
				<para>Extension number</para>
			</parameter>
		</syntax>
		<description>
			 <para>Gets or sets the authentication pin given a extension number.</para>
		</description>
	</function>

	<function name="USER_PIN" language="en_US">
		<synopsis>
			Check or modify user authentication pin
		</synopsis>
		<syntax>
			<parameter name="pin" required="true">
				<para>User's pin number</para>
			</parameter>
		</syntax>
		<description>
			 <para>Gets or sets user's authentication pin</para>
			 <para>Sets variable <variable>RESULT</variable> to <literal>1</literal> on success, <literal>-1</literal> on database error or <literal>0</literal> on failure (user pin not modified or it doesn't exist)</para>
		</description>
	</function>

	<function name="CELL_EXTEN" language="en_US">
		<synopsis>
			Gets a extension number given a cellular number
		</synopsis>
		<syntax>
			<parameter name="exten" required="false">
				<para>Extension number (optional)</para>
			</parameter>
		</syntax>
		<description>
			 <para>Gets or sets the authentication pin given a extension number.</para>
			 <para>If no extension given, use current channel callerid</para>
		</description>
	</function>

	<function name="CALLERID_CELLEXTEN" language="en_US">
		<synopsis>
			Returns callerid or corresponding exten number
		</synopsis>
		<syntax>
			<parameter name="exten" required="false">
				<para>Extension number</para>
			</parameter>
		</syntax>
		<description>
			 <para>Returns <variable>CALLERID</variable> of channel if variable <variable>INCOMING_CELL</variable> is not set, or returns <variable>CELL_EXTEN(${CALLERID(num)}</variable> if variable is set</para>
		</description>
	</function>
	
	<function name="IPC_MYSQL" language="en_US">
		<synopsis>
			Executes an arbitrary SQL query an returns result
		</synopsis>
		<syntax>
			<parameter name="sql" required="true">
				<para>SQL query string</para>
			</parameter>
		</syntax>
		<description>
			 <para>Executes an SQL query in IPContact DB and returns the result. If query returns more than one row, only first row is returned; if it returns multiple fields, only first field is returned.</para>
			 <para>Only SELECT queries are allowed for security</para>
		</description>
	</function>

***/

static struct ast_ipcontact_conf ipcontact = {
	.dbhost = "",
	.dbname = "",
	.dbuser = "",
	.dbpass = "",
	.dbport = 0,
	.dbsock = "",
	.tables = "",
};

static struct virtual_pbx default_vpbx = {
	.number = 0,
	.family = 0,
	.phone = "",
	.name = "Default PBX",
	.mainmenu = "mainmenu" AST_STANDARD_APP_DELIM "s" AST_STANDARD_APP_DELIM "1",
	.operator = "colas" AST_STANDARD_APP_DELIM "operadora" AST_STANDARD_APP_DELIM "1",
	.vmslocal = "vmselect_local",
	.vmsincoming = "vmselect_extern",
	.dialoptsloc = "30" AST_STANDARD_APP_DELIM "tTr",
	.dialoptsext = "30" AST_STANDARD_APP_DELIM "tm",
	.dialoptsret = "20" AST_STANDARD_APP_DELIM "tm",
	.monitor = 1,
};

static AST_LIST_HEAD_STATIC(vpbxs, virtual_pbx);
static int vpbx_count = 0;

static AST_LIST_HEAD_STATIC(tables, ast_ipcontact_table);
static int table_count = 0;


/* Globals */
static const char config_file[] = "ipcontact.conf";
static const char db_root[] = "ipcontact";
static int debug = 0;
int schemaversion = 1;

/* Global MySQL variables */
static MYSQL mysql;
static int connected = 0;


AST_MUTEX_DEFINE_STATIC(ipc_lock);

/**** Virtual PBXs functions ****/
static struct virtual_pbx *make_default_vpbx(void);
inline static void flush_vpbxs(void);
static int reload_vpbxs(void);
static int astdb_read_vpbxs(void);
static void astdb_write_vpbxs(void);
static int reload_exten_info(void);
static int reload_agent_info(void);



inline int ipc_mutex_lock()
{
	return ast_mutex_lock(&ipc_lock);
}

inline int ipc_mutex_unlock()
{
	return ast_mutex_unlock(&ipc_lock);
}



inline const char *ast_channel_type(struct ast_channel *chan)
{
	if (chan->tech && chan->tech->type) {
		return chan->tech->type;
	} else {
		return NULL;
	}
}


/*
 * Help for commands provided by this module ...
 */
static char help_reload_config[] =
	"Usage: ipcontact reload config\n"
	"       (Re)load the ipcontact system configuration into local Asterisk DB\n";

static char help_reload_extensions[] =
	"Usage: ipcontact reload vpbxs\n"
	"       (Re)load extensions configuration info into local Asterisk DB\n";

static char help_reload_vpbxs[] =
	"Usage: ipcontact reload vpbxs"
	"       (Re)load virtual pbx list into local Asterisk IPContact Module\n";

static char help_reload_agents[] =
	"Usage: ipcontact reload agents\n"
	"       (Re)load agents configuration info into local Asterisk DB\n";

static char help_reload_tables[] =
	"Usage: ipcontact reload tables\n"
	"       (Re)load custom tables configuration info into local Asterisk DB\n";

static char help_show_vpbxs[] =
	"Usage: ipcontact show vpbxs" "       Show IPContact current list of virtual pbxs\n";

static char help_show_version[] =
	"Usage: ipcontact show version" "      Show IPContact Addons version\n";

static char help_show_config[] =
	"Usage: ipcontact show config" "      Show IPContact config parameters\n";


int is_number(const char *num)
{
	int i = 0, ret = 1;

	if (!num || ast_strlen_zero(num))
		return 0;
	for (i = 1; i < strlen(num); i++) {
		if (!isdigit(num[i])) {
			ret = 0;
			break;
		}
	}
	return ret;
}



static struct virtual_pbx *make_default_vpbx(void)
{
	struct virtual_pbx *tmp = (struct virtual_pbx *) malloc(sizeof(struct virtual_pbx));
	if (!tmp) {
		ast_log(LOG_ERROR, "Unable to MALLOC default pbx\n");
		return NULL;
	}
	memset(tmp, 0, sizeof(struct virtual_pbx));
	ast_copy_string(tmp->phone, "", sizeof(tmp->phone));
	ast_copy_string(tmp->name, "Default PBX", sizeof(tmp->name));
	ast_copy_string(tmp->mainmenu, default_vpbx.mainmenu, sizeof(tmp->mainmenu));
	ast_copy_string(tmp->operator, default_vpbx.operator, sizeof(tmp->operator));
	ast_copy_string(tmp->vmslocal, default_vpbx.vmslocal, sizeof(tmp->vmslocal));
	ast_copy_string(tmp->vmsincoming, default_vpbx.vmsincoming, sizeof(tmp->vmsincoming));
	ast_copy_string(tmp->dialoptsloc, default_vpbx.dialoptsloc, sizeof(tmp->dialoptsloc));
	ast_copy_string(tmp->dialoptsext, default_vpbx.dialoptsext, sizeof(tmp->dialoptsext));
	ast_copy_string(tmp->dialoptsret, default_vpbx.dialoptsret, sizeof(tmp->dialoptsret));
	return tmp;
}

static void flush_vpbxs(void)
{
	struct virtual_pbx *tmp;

	AST_LIST_LOCK(&vpbxs);
	while ((tmp = AST_LIST_REMOVE_HEAD(&vpbxs, list))) {
		free(tmp);
	}
	vpbx_count = 0;
	AST_LIST_UNLOCK(&vpbxs);
}


static void flush_tables(void)
{
	struct ast_ipcontact_table *tmp;

	AST_LIST_LOCK(&tables);
	while ((tmp = AST_LIST_REMOVE_HEAD(&tables, list))) {
		free(tmp);
	}
	table_count = 0;
	AST_LIST_UNLOCK(&tables);
}


/*!
 * \brief Initialize MySQL handler and connect to DB.
 * \return pointer to MYSQL connection on success or NULL if unable to connect
 */
MYSQL *ipc_sql_reconnect(void)
{
	static time_t connect_time = 0, ctime = 0;
	static unsigned long thread_id;
	
	/* Global variables: mysql, connected */

	if (!connected) {
		/* Attempt to connect */
		if(!strncmp(ipcontact.dbhost, "localhost", 9)) {
			ast_verbose(VERBOSE_PREFIX_3
									"Connecting to MySQL server on %s with user %s, password %s, database %s, using UNIX socket %s\n",
									ipcontact.dbhost, ipcontact.dbuser, ipcontact.dbpass, ipcontact.dbname, ipcontact.dbsock);
		} else {
			ast_verbose(VERBOSE_PREFIX_3
									"Connecting to MySQL server %s:%u with user %s, password %s, database %s\n",
									ipcontact.dbhost, ipcontact.dbport, ipcontact.dbuser, ipcontact.dbpass, ipcontact.dbname);
		}
		mysql_init(&mysql);
		/* Automatically reconnect on failure (mysql_ping) */
		mysql_options(&mysql, MYSQL_OPT_RECONNECT, "1");
		if (mysql_real_connect(&mysql, ipcontact.dbhost[0] ? ipcontact.dbhost : NULL,
													 ipcontact.dbuser, ipcontact.dbpass, ipcontact.dbname, ipcontact.dbport,
													 ipcontact.dbsock[0] ? ipcontact.dbsock : NULL, 0)) {
			thread_id = mysql_thread_id(&mysql);
			ast_log(LOG_DEBUG, "Successfully connected to MySQL database, connection id %lu\n", thread_id);
			connect_time = time(NULL);
			connected = 1;
		} else {
			ast_log(LOG_ERROR, "Failed to connect to database: Error (%i): %s\n",
							mysql_errno(&mysql), mysql_error(&mysql));
			connected = 0;
			ipc_sql_disconnect();
			return NULL;
		}
	} else {
		ctime = time(NULL) - connect_time;

		if (ctime > IPC_MYSQL_PING_TIME) {		/* Long connection - ping the server */
			if (debug)
				ast_log(LOG_NOTICE, "Mysql connection: pinging the server\n");
			if (mysql_ping(&mysql)) {
				/* Get here only on fatal error, MYSQL_OPT_RECONNECT is true */
				ast_log(LOG_WARNING,"Error (%i): %s\n", mysql_errno(&mysql), mysql_error(&mysql));
				connected = 0;
				ipc_sql_disconnect();
				return NULL;
			} else {
				if (mysql_thread_id(&mysql) != thread_id) {
					thread_id = mysql_thread_id(&mysql);
					if (debug)
						ast_log(LOG_NOTICE, "Mysql connection: reconnected to server, connection id: %lu\n", thread_id);
				} else {
					if (debug)
						ast_log(LOG_NOTICE, "Mysql connection: server is ALIVE\n");
				}
				connect_time = time(NULL);
			}
		}
	}
	return &mysql;
}

/*! \brief Disconnect from MySQL DB */
void ipc_sql_disconnect(void)
{
	if (connected) {
		mysql_close(&mysql);
		connected = 0;
	}
}

/*!
 * \brief Execute a MySQL query and returns MySQL result
 * \return MYSQL_RES on success or NULL on error
 */
MYSQL_RES *ipc_sql_exec_query(char *sql)
{
	MYSQL *conn;
	MYSQL_RES *sqlres = NULL;

	ast_mutex_lock(&ipc_lock);
	conn = ipc_sql_reconnect();
	if (!conn) {
		ast_log(LOG_ERROR, "Unable to connect to MySQL server\n");
		ast_mutex_unlock(&ipc_lock);
		return NULL;
	};

	if (debug) {
		ast_log(LOG_NOTICE, "Executing MySQL Query: %s\n", sql);
	}

	if (mysql_real_query(conn, sql, strlen(sql))) {
		ast_log(LOG_ERROR, "IPContact: Failed to query database.\n");
		ast_log(LOG_ERROR, "IPContact: Query: %s\n", sql);
		ast_log(LOG_ERROR, "IPContact: MySQL error message: %s\n", mysql_error(conn));
		ast_mutex_unlock(&ipc_lock);
		return NULL;
	}

	sqlres = mysql_store_result(conn);

	if (sqlres == NULL) {
		ast_log(LOG_ERROR, "Result error: %s\n", mysql_error(conn));
	}
	ast_mutex_unlock(&ipc_lock);
	return sqlres;
}


/* Supported formats:
   1X{1,6}
   [3-9]XX XXXX, 2 XXX XXXX, 02 XXX XXXX
   9[4-9] XXXXXX, 09[4-9] XXXXXX, 09 XXXXXX
   0[1-8]XXXXXX.
   00XXXX.
*/
#if defined(HAS_AST16) || defined(HAS_AST18)
static int normantel_read(struct ast_channel *chan, const char *cmd, char *data, char *buf, size_t len)
#else
static int normantel_read(struct ast_channel *chan, char *cmd, char *data, char *buf, size_t len)
#endif
{
	char *target = buf;
	const char *dataptr = NULL;
	int res = 0;
	int numlen = 0;
	enum { CID_LOCAL, CID_CELL, CID_FULL } cid_type;

	AST_DECLARE_APP_ARGS(args, AST_APP_ARG(cid););
	AST_STANDARD_APP_ARGS(args, data);

	if (!args.cid) {
		return RESULT_SHOWUSAGE;
	}
	numlen = strlen(args.cid);
	if (numlen < 2 || !is_number(args.cid)) {
		ast_copy_string(target, args.cid, len);
		return 0;
	}
	dataptr = args.cid;
	if (*dataptr == '0') {
		if (*(dataptr + 1) == '9' && (numlen == 9 || numlen == 8) && *(dataptr + 2) != '0') {	/* 09N XXXXXX */
			dataptr++;
			numlen--;
			cid_type = CID_CELL;
		} else if (*(dataptr + 1) == '2') {	/* 02 XXXXXXX */
			dataptr = dataptr + 2;
			numlen = numlen - 2;
			cid_type = CID_LOCAL;
		} else {				/* 090XXXXXXX, 0[13-8]XXXXXXX */
			cid_type = CID_FULL;
		}
	} else if (*dataptr == '2' && numlen == 8) {	/* 2XXXXXXX */
		dataptr++;
		numlen--;
		cid_type = CID_LOCAL;
	} else if ((*dataptr == '9') && (*(dataptr + 1) != '0') && (numlen == 8)) {	/* 99 XXXXXX */
		cid_type = CID_CELL;
	} else {
		cid_type = CID_LOCAL;
	}

	switch (cid_type) {
	case CID_LOCAL:
		snprintf(target, len, "02%s", dataptr);
		break;
	case CID_CELL:
		if (numlen == 7) {		/* 09 XXXXXX: es ANCEL */
			snprintf(target, len, "09%s", dataptr);
		} else if (numlen == 8) {	/* 9X XXXXXX */
			dataptr++;
			if ((*dataptr) == '1' || (*dataptr) == '2' || (*dataptr) == '3') {
				ast_log(LOG_WARNING, "Unknown cellular number: %s\n", args.cid);
			}
			snprintf(target, len, "09%s", dataptr);
		}
		break;
	case CID_FULL:
		ast_copy_string(target, dataptr, len);
		break;
	default:
		ast_log(LOG_WARNING, "Invalid CID number pattern: %s\n", args.cid);
		ast_copy_string(target, dataptr, len);
	}

	if ((res == 0) && debug) {
		ast_verbose("%s -> parsed as -> %s\n", args.cid, target);
	}
	// Make sure output string is null-terminated 
	target[len - 1] = '\0';
	return 0;
}

#if defined(HAS_AST16) || defined(HAS_AST18)
static int profileauth_read(struct ast_channel *chan, const char *cmd, char *data, char *buf, size_t len)
#else
static int profileauth_read(struct ast_channel *chan, char *cmd, char *data, char *buf, size_t len)
#endif
{

	int res;
	char calltype[16] = IPC_PROFILE_LOCAL;

	AST_DECLARE_APP_ARGS(args, AST_APP_ARG(opt););
	AST_STANDARD_APP_ARGS(args, data);

	buf[0] = '\0';
	if (args.opt) {
		if (!strncasecmp(args.opt, "loc", 3)) {	/* Local call */
			ast_copy_string(calltype, IPC_PROFILE_LOCAL, sizeof(calltype));
		} else if (!strncasecmp(args.opt, "cel", 3)) {	/* Cellular call */
			ast_copy_string(calltype, IPC_PROFILE_CELLULAR, sizeof(calltype));
		} else if (!strncasecmp(args.opt, "lon", 3)) {	/* Long Distance call */
			ast_copy_string(calltype, IPC_PROFILE_LONGDIST, sizeof(calltype));
		} else if (!strncasecmp(args.opt, "int", 3)) {	/* International call */
			ast_copy_string(calltype, IPC_PROFILE_INTER, sizeof(calltype));
		} else {
			return RESULT_SHOWUSAGE;
		}
	}


	if (ast_strlen_zero(CALLERID_NUM(chan))) {
		ast_log(LOG_WARNING, "CallerID not set on channel: %s. Unable to get profile\n",
				chan->name);
		return 0;
	}

	res = get_exten_profile(CALLERID_NUM(chan), calltype);

	if (res >= 0) {
		ast_copy_string(buf, res ? "1" : "0", len);
	}

	return 0;
}


int get_exten_profile(const char *exten, const char *calltype)
{

	char dbres[16] = "";
	char tree[IPC_MAX_LEN] = "";

	if (ast_strlen_zero(exten)) {
		ast_log(LOG_WARNING, "Unable to get profile for empty extension!\n");
		return -1;
	}
	if (ast_strlen_zero(calltype)) {
		ast_log(LOG_WARNING,
				"Trying to get profile for unknown call type on extension %s!\n", exten);
		return -1;
	}

	snprintf(tree, sizeof(tree), "%s/%s/profile", IPC_EXTEN_TREE, exten);
	if (!ast_db_get(tree, calltype, dbres, sizeof(dbres) - 1)) {
		return ast_true(dbres) ? 1 : 0;
	} else {
		ast_log(LOG_WARNING, "Profile not found for calltype %s on extension %s\n", calltype, exten);
		return -1;
	}
}

#if defined(HAS_AST16) || defined(HAS_AST18)
static int extenpin_read(struct ast_channel *chan, const char *cmd, char *data, char *buf, size_t len)
#else
static int extenpin_read(struct ast_channel *chan, char *cmd, char *data, char *buf, size_t len)
#endif
{

	int res = 0;
	MYSQL *conn = NULL;
	MYSQL_ROW row;
	MYSQL_RES *sqlres = NULL;
	char query[IPC_QUERY_LEN];

	if (ast_strlen_zero(data)) {
		ast_log(LOG_ERROR, "Must specify extension to read!\n");
		pbx_builtin_setvar_helper(chan, "RESULT", "");
		res = -1;
	} else {
		/**** BEGIN MUTEX BLOCK ****/
		ipc_mutex_lock();
		conn = ipc_sql_reconnect();
		if (!conn) {
			ast_log(LOG_ERROR, "Unable to connect to MySQL server\n");
			pbx_builtin_setvar_helper(chan, "RESULT", "-1");
			res = -1;
		} else {
			/* Leo el valor de pin para la extension */
			snprintf(query, sizeof(query),
					 "SELECT pin FROM extension WHERE number=\'%s\'", data);
			if (debug) {
				ast_log(LOG_NOTICE, "Executing MySQL Query: %s\n", query);
			}
			if (mysql_real_query(conn, query, strlen(query))) {
				ast_log(LOG_ERROR, "Database query failed: %s.\n", query);
				ast_log(LOG_ERROR, "MySQL error message: %s\n", mysql_error(conn));
				pbx_builtin_setvar_helper(chan, "RESULT", "-1");
				res = -1;
			} else {
				/* Parseo el resultado */
				if ((sqlres = mysql_store_result(conn))) {
					if (mysql_num_rows(sqlres) > 0) {
						/* row nunca puee ser null, pin es clave */
						row = mysql_fetch_row(sqlres);
						snprintf(buf, len, row[0]);
						pbx_builtin_setvar_helper(chan, "RESULT", "1");
					} else {
						ast_log(LOG_WARNING, "Could not find desired extension\n");
						pbx_builtin_setvar_helper(chan, "RESULT", "0");
					}
					mysql_free_result(sqlres);
					sqlres = NULL;
				} else {
					ast_log(LOG_ERROR, "MySQL error: ERROR %i (%s)\n", mysql_errno(conn), mysql_error(conn));
					pbx_builtin_setvar_helper(chan, "RESULT", "-1");
					res = -1;
				}
			}
		}
		/**** END MUTEX BLOCK ****/
		ipc_mutex_unlock();
	}
	return res;
}

#if defined(HAS_AST16) || defined(HAS_AST18)
static int extenpin_write(struct ast_channel *chan, const char *cmd, char *data, const char *value)
#else
static int extenpin_write(struct ast_channel *chan, char *cmd, char *data, const char *value)
#endif
{

	int res = 0;
	MYSQL *conn = NULL;
	char query[IPC_QUERY_LEN];
	char pin[10] = "";

	if (ast_strlen_zero(data)) {
		ast_log(LOG_ERROR, "Must specify extension to write!\n");
		res = -1;
	} else {
		if (ast_strlen_zero(value)) {
			ast_log(LOG_ERROR, "Must specify a pin to write!\n");
			res = -1;
		} else {
			ast_copy_string(pin, value, 10);
		}
	}

	if (res) {
		pbx_builtin_setvar_helper(chan, "RESULT", "");
	} else {
		/**** BEGIN MUTEX BLOCK ****/
		ipc_mutex_lock();
		conn = ipc_sql_reconnect();
		if (!conn) {
			ast_log(LOG_ERROR, "Unable to connect to MySQL server\n");
			pbx_builtin_setvar_helper(chan, "RESULT", "-1");
			res = -1;
		} else {
			/* Actualizo valor de pin para la extension */
			snprintf(query, sizeof(query),
					 "UPDATE extension SET pin='%s' WHERE number=\'%s\'", pin, data);
			if (debug) {
				ast_log(LOG_NOTICE, "Executing MySQL Query: %s\n", query);
			}
			if (mysql_real_query(conn, query, strlen(query))) {
				ast_log(LOG_ERROR, "Database query failed: %s.\n", query);
				ast_log(LOG_ERROR, "MySQL error message: %s\n", mysql_error(conn));
				pbx_builtin_setvar_helper(chan, "RESULT", "-1");
				res = -1;
			} else {
				if (mysql_affected_rows(conn) > 0) {
					pbx_builtin_setvar_helper(chan, "RESULT", "1");
				} else {
					ast_log(LOG_WARNING, "No rows affected, no extension modified\n");
					pbx_builtin_setvar_helper(chan, "RESULT", "0");
				}
			}
		}
		/**** END MUTEX BLOCK ****/
		ipc_mutex_unlock();
	}
	return res;
}

#if defined(HAS_AST16) || defined(HAS_AST18)
static int userpin_read(struct ast_channel *chan, const char *cmd, char *data, char *buf, size_t len)
#else
static int userpin_read(struct ast_channel *chan, char *cmd, char *data, char *buf, size_t len)
#endif
{

	int res = 0;
	MYSQL *conn = NULL;
	MYSQL_ROW row;
	MYSQL_RES *sqlres = NULL;
	char query[IPC_QUERY_LEN];

	if (ast_strlen_zero(data)) {
		ast_log(LOG_ERROR, "Must specify pin to check!\n");
		pbx_builtin_setvar_helper(chan, "RESULT", "");
		res = -1;
	} else {
		/**** BEGIN MUTEX BLOCK ****/
		ipc_mutex_lock();
		conn = ipc_sql_reconnect();
		if (!conn) {
			ast_log(LOG_ERROR, "Unable to connect to MySQL server\n");
			pbx_builtin_setvar_helper(chan, "RESULT", "-1");
			res = -1;
		} else {
			/* Leo el valor de pin para la extension */
			if (schemaversion == 1) {
				snprintf(query, sizeof(query),
								 "SELECT 1 FROM authcalls WHERE pin='%s' AND enabled", data);
			}	else {
				snprintf(query, sizeof(query),
								 "SELECT 1 FROM authcalls a, users u WHERE pin='%s' AND enabled AND a.user_id=u.user_id", data);
			}
			if (debug) {
				ast_log(LOG_NOTICE, "Executing MySQL Query: %s\n", query);
			}
			if (mysql_real_query(conn, query, strlen(query))) {
				ast_log(LOG_ERROR, "Database query failed: %s.\n", query);
				ast_log(LOG_ERROR, "MySQL error message: %s\n", mysql_error(conn));
				pbx_builtin_setvar_helper(chan, "RESULT", "-1");
				res = -1;
			} else {
				/* Parseo el resultado */
				if ((sqlres = mysql_store_result(conn))) {
					if (mysql_num_rows(sqlres) > 0) {
						/* row nunca puee ser null, pin es clave */
						row = mysql_fetch_row(sqlres);
						snprintf(buf, len, row[0]);
						pbx_builtin_setvar_helper(chan, "RESULT", "1");
					} else {
						pbx_builtin_setvar_helper(chan, "RESULT", "0");
					}
					mysql_free_result(sqlres);
					sqlres = NULL;
				} else {
					ast_log(LOG_ERROR, "MySQL error: ERROR %i (%s)\n", mysql_errno(conn), mysql_error(conn));
					pbx_builtin_setvar_helper(chan, "RESULT", "-1");
					res = -1;
				}
			}
		}
		/**** END MUTEX BLOCK ****/
		ipc_mutex_unlock();
	}
	return res;
}

#if defined(HAS_AST16) || defined(HAS_AST18)
static int userpin_write(struct ast_channel *chan, const char *cmd, char *data, const char *value)
#else
static int userpin_write(struct ast_channel *chan, char *cmd, char *data, const char *value)
#endif
{

	int res = 0;
	MYSQL *conn = NULL;
	char query[IPC_QUERY_LEN];
	char pin[10] = "";

	if (ast_strlen_zero(data)) {
		ast_log(LOG_ERROR, "Must specify original pin!\n");
		res = -1;
	} else {
		if (ast_strlen_zero(value)) {
			ast_log(LOG_ERROR, "Must specify a pin to write!\n");
			res = -1;
		} else {
			ast_copy_string(pin, value, 10);
		}
	}

	if (res) {
		pbx_builtin_setvar_helper(chan, "RESULT", "");
	} else {
		/**** BEGIN MUTEX BLOCK ****/
		ipc_mutex_lock();
		conn = ipc_sql_reconnect();
		if (!conn) {
			ast_log(LOG_ERROR, "Unable to connect to MySQL server\n");
			pbx_builtin_setvar_helper(chan, "RESULT", "-1");
			res = -1;
		} else {
			/* Actualizo valor de pin para la extension */
			snprintf(query, sizeof(query),
					 "UPDATE authcalls SET pin='%s' WHERE pin=\'%s\'", value, data);
			if (debug) {
				ast_log(LOG_NOTICE, "Executing MySQL Query: %s\n", query);
			}
			if (mysql_real_query(conn, query, strlen(query))) {
				ast_log(LOG_ERROR, "Database query failed: %s.\n", query);
				ast_log(LOG_ERROR, "MySQL error message: %s\n", mysql_error(conn));
				pbx_builtin_setvar_helper(chan, "RESULT", "-1");
				res = -1;
			} else {
				if (mysql_affected_rows(conn) > 0) {
					pbx_builtin_setvar_helper(chan, "RESULT", "1");
				} else {
					ast_log(LOG_WARNING, "No rows affected, no pin modified\n");
					pbx_builtin_setvar_helper(chan, "RESULT", "0");
				}
			}
		}
		/**** END MUTEX BLOCK ****/
		ipc_mutex_unlock();
	}
	return res;
}

#if defined(HAS_AST16) || defined(HAS_AST18)
static int cellexten_read(struct ast_channel *chan, const char *cmd, char *data, char *buf, size_t len)
#else
static int cellexten_read(struct ast_channel *chan, char *cmd, char *data, char *buf, size_t len)
#endif
{
  int res = 0;
  MYSQL *conn = NULL;
  MYSQL_ROW row;
  MYSQL_RES *sqlres = NULL;
  char query[IPC_QUERY_LEN];
  char cell[IPC_MAX_LEN] = "";
				
	if (ast_strlen_zero(data)) {
		if (debug) {
			ast_log(LOG_NOTICE, "Called with no args... Using channel's callerid number: %s\n", CALLERID_NUM(chan));
		}
		strncpy(cell, CALLERID_NUM(chan), IPC_MAX_LEN);
	} else {
		strncpy(cell, data, IPC_MAX_LEN);
	}


	/**** BEGIN MUTEX BLOCK ****/
	ipc_mutex_lock();

	conn = ipc_sql_reconnect();
	if (!conn) {
		ast_log(LOG_ERROR, "Unable to connect to MySQL server\n");
		pbx_builtin_setvar_helper(chan, "RESULT", "-1");
		res = -1;
	} else {
		/* Leo el valor que tiene asociado el celular al interno */
		snprintf(query, sizeof(query),"SELECT number FROM extension WHERE fwmcellular=\'%s\'", cell);
		if(debug) {
			ast_log(LOG_NOTICE, "Executing MySQL Query: %s\n", query);
		}
		if (mysql_real_query(conn, query, strlen(query))) {
			ast_log(LOG_ERROR, "Database query failed: %s.\n",query);
			ast_log(LOG_ERROR, "MySQL error message: %s\n", mysql_error(conn));
			pbx_builtin_setvar_helper(chan, "RESULT", "-1");
			res = -1;
		} else {
			/* Parseo el resultado */
			if ((sqlres = mysql_store_result(conn))) {
				if(mysql_num_rows(sqlres) > 0) {
					/* row puede ser null */
					row = mysql_fetch_row(sqlres);
					snprintf(buf, len, row[0]);
					pbx_builtin_setvar_helper(chan, "RESULT", "1");
				} else {
					pbx_builtin_setvar_helper(chan, "RESULT", "0");
				}
				mysql_free_result(sqlres);
				sqlres = NULL;
			} else {
				ast_log(LOG_ERROR, "MySQL error: ERROR %i (%s)\n", mysql_errno(conn), mysql_error(conn));
				pbx_builtin_setvar_helper(chan, "RESULT", "-1");
				res = -1;
			}
		}
	}
	/**** END MUTEX BLOCK ****/
	ipc_mutex_unlock();

	return res;
}

#if defined(HAS_AST16) || defined(HAS_AST18)
static int calleridcellexten_read(struct ast_channel *chan, const char *cmd, char *data, char *buf, size_t len)
#else
static int calleridcellexten_read(struct ast_channel *chan, char *cmd, char *data, char *buf, size_t len)
#endif
{
  int res = 0;

	if (IS_INCOMINGCELL_CALL(chan)) {
		if (debug) {
			ast_log(LOG_DEBUG, "Incoming cell call\n");
		}
		res = cellexten_read(chan, cmd, data, buf, len);
	} else {
		if (debug) {
			ast_log(LOG_DEBUG, "Normal call\n");
		}
		strncpy(buf, CALLERID_NUM(chan), len);
		pbx_builtin_setvar_helper(chan, "RESULT", "1");
	}
	return res;
}

#if defined(HAS_AST16) || defined(HAS_AST18)
static int ipcmysql_read(struct ast_channel *chan, const char *cmd, char *data, char *buf, size_t len)
#else
static int ipcmysql_read(struct ast_channel *chan, char *cmd, char *data, char *buf, size_t len)
#endif
{
	int res = 0;
	MYSQL *conn = NULL;
	MYSQL_ROW row;
	MYSQL_RES *sqlres = NULL;
	char query[IPC_QUERY_LEN];

	if (ast_strlen_zero(data)) {
		ast_log(LOG_ERROR, "Must specify MySQL query!\n");
		pbx_builtin_setvar_helper(chan, "RESULT", "");
		res = -1;
	} else {
		/* Safety check, only select queries allowed */
		if (!strcasestr(data, "select")) {
			ast_log(LOG_ERROR, "Only SELECT queries are allowed!\n");
			return -1;
		}

		/**** BEGIN MUTEX BLOCK ****/
		ipc_mutex_lock();
		conn = ipc_sql_reconnect();
		if (!conn) {
			ast_log(LOG_ERROR, "Unable to connect to MySQL server\n");
			pbx_builtin_setvar_helper(chan, "RESULT", "-1");
			res = -1;
		} else {
			ast_copy_string(query, data, sizeof(query));
			if (debug) {
				ast_log(LOG_NOTICE, "Executing MySQL Query: %s\n", query);
			}
			if (mysql_real_query(conn, query, strlen(query))) {
				ast_log(LOG_ERROR, "Database query failed: %s.\n", query);
				ast_log(LOG_ERROR, "MySQL error message: %s\n", mysql_error(conn));
				pbx_builtin_setvar_helper(chan, "RESULT", "-1");
				res = -1;
			} else {
				/* Parseo el resultado */
				if ((sqlres = mysql_store_result(conn))) {
					if (mysql_num_rows(sqlres) > 0) {
						/* hay al menos 1 fila */
						row = mysql_fetch_row(sqlres);
						snprintf(buf, len, row[0]);
						pbx_builtin_setvar_helper(chan, "RESULT", "1");
					} else {
						ast_log(LOG_WARNING, "Query returned empty resultset\n");
						pbx_builtin_setvar_helper(chan, "RESULT", "0");
					}
					mysql_free_result(sqlres);
					sqlres = NULL;
				} else {
					ast_log(LOG_WARNING,
									"MySQL error: ERROR %i (%s)\n", mysql_errno(conn), mysql_error(conn));
					pbx_builtin_setvar_helper(chan, "RESULT", "-1");
					res = -1;
				}
			}
		}
		/**** END MUTEX BLOCK ****/
		ipc_mutex_unlock();
	}
	return res;
}

/* Functions */
static struct ast_custom_function normantel_function = {
	.name = "NORM_ANTEL",
#ifndef HAS_XMLDOC
	.synopsis = "Normalize ANTEL callerid numbers",
	.syntax = "NORM_ANTEL(<cidnumber>)",
	.desc =
	"Normalize a callerid number as received from ANTEL.\n"
	"Returned number must follow the pattern:\n"
	"[local code]XXXX for national calls\n"
	"09XXXXXXX for cellular numbers\n"
	"1XX. , 0800XXXX and 0900XXXX are not changed\n",
#endif
	.read = normantel_read,
};

static struct ast_custom_function profileauth_function = {
	.name = "PROFILE_AUTH",
#ifndef HAS_XMLDOC
	.synopsis = "Check profile auth to make given call type from current extension",
	.syntax = "PROFILE_AUTH(<type>)",
	.desc =
	"Check if the curren extension has permissions to initiate a call of the given type.\n"
	"Valid call types values are:\n"
	"- local (default)   PSTN Local calls.\n"
	"- cell              Cellular calls.\n"
	"- long              Long distance calls.\n"
	"- inter             International calls.\n",
#endif
	.read = profileauth_read,
};

static struct ast_custom_function extenpin_function = {
	.name = "EXTEN_PIN",
#ifndef HAS_XMLDOC
	.synopsis = "Gets or sets extension's authentication pin\n",
	.syntax = "EXTEN_PIN(<exten>)",
	.desc = "Gets or sets the authentication pin given a extension number.\n",
#endif
	.read = extenpin_read,
	.write = extenpin_write,
};

static struct ast_custom_function userpin_function = {
	.name = "USER_PIN",
#ifndef HAS_XMLDOC
	.synopsis = "Check or modify user authentication pin\n",
	.syntax = "USER_PIN(<pin>)",
	.desc = "Gets or sets user's authentication pin.\nSets variable RESULT to 1 on success, -1 on database error or 0 on failure (user pin not modified or it doesn't exist)\n",
#endif
	.read = userpin_read,
	.write = userpin_write,
};

static struct ast_custom_function cellexten_function = {
	.name = "CELL_EXTEN",
#ifndef HAS_XMLDOC
	.synopsis = "Gets a extension number given a cellular number.\n",
	.syntax = "CELL_EXTEN(<cell_number>)",
	.desc = "Gets extension number given a cellular number associated with it via extension table.\n",
#endif
	.read = cellexten_read,
};
 
static struct ast_custom_function calleridcellexten_function = {
	.name = "CALLERID_CELLEXTEN",
#ifndef HAS_XMLDOC
	.synopsis = "Returns callerid or corresponding exten number.\n",
	.syntax = "CALLERID_CELLEXTEN()",
	.desc = "Returns CALLERID of channel if variable INCOMING_CELL is not set or returns CELL_EXTEN(${CALLERID(num)) if variable is set.\n",
#endif
	.read = calleridcellexten_read,
};

static struct ast_custom_function ipcmysql_function = {
	.name = "IPC_MYSQL",
#ifndef HAS_XMLDOC
	.synopsis = "Executes an arbitrary SQL query an returns result.\n",
	.syntax = "IPC_MYSQL(sql)",
	.desc = "Executes an SQL query in IPContact DB and returns the result. If query returns more than one row, only first row is returned; if it returns multiple fields, only first field is returned. Only SELECT queries are allowed for security.\n",
#endif
	.read = ipcmysql_read,
};

/*!
 * \brief Reload the extensions info from MYSQL into Asterisk DB
 * \return 0 on success or -1 if unable to connect or query fails
 */
static int reload_exten_info(void)
{
	MYSQL *conn;
	MYSQL_RES *sqlres = NULL;
	MYSQL_FIELD *fields;
	MYSQL_ROW row;
	int num_fields;
	char exten_tree[128];
	char profile_tree[128];
	int i, res = 0;
	char sql[IPC_QUERYLONG_LEN];

	if (schemaversion == 1) {
		snprintf(sql, sizeof(sql), "SELECT * FROM extension");
	} else {
		snprintf(sql, sizeof(sql),
						 "SELECT number,e.vpbx,info, v.mailbox AS voicemail, if(e.protocol is null, p.protocol, e.protocol) AS protocol, channel, if(e.vm_flag is null, p.vm_flag, e.vm_flag) AS vm_flag, pin, if(e.monitor is null, p.monitor, e.monitor) AS monitor, fwmcellular, if(e.time_follow is null, p.time_follow, e.time_follow) AS time_follow, if(e.local is null, p.local, e.local) AS local, if(e.longdist is null, p.longdist, e.longdist) AS longdist, if (e.cellular is null, p.cellular, e.cellular) AS cellular, if(e.inter is null, p.inter, e.inter) AS inter, if(e.time_cell is null, p.time_cell, e.time_cell) AS time_cell, fwmskype FROM extension e LEFT JOIN pattern p ON (e.pattern = p.id) LEFT JOIN voicemail_account v ON (e.voicemail = v.uniqueid)");
	}

	/***** BEGIN MUTEX BLOCK **********/
	ast_mutex_lock(&ipc_lock);

	conn = ipc_sql_reconnect();
	if (!conn) {
		ast_log(LOG_ERROR, "Unable to connect to MySQL server\n");
		ast_mutex_unlock(&ipc_lock);
		return -1;
	};

	/* Execution. */
	if (mysql_real_query(conn, sql, strlen(sql))) {
		ast_log(LOG_ERROR, "IPContact reload extensions: Failed to query database.\n");
		ast_log(LOG_ERROR, "IPContact reload extensions: Query: %s\n", sql);
		ast_log(LOG_ERROR, "IPContact reload extensions: MySQL error message: %s\n",
				mysql_error(conn));
		ast_mutex_unlock(&ipc_lock);
		return -1;
	}

	if ((sqlres = mysql_store_result(conn))) {
		/* Borro todo el arbol de extensiones antes de empezar siempre*/
		ast_db_deltree("ipcontact/extension", 0);

		if (mysql_num_rows(sqlres) > 0) {
			num_fields = mysql_num_fields(sqlres);
			fields = mysql_fetch_fields(sqlres);

			while ((row = mysql_fetch_row(sqlres))) {
				snprintf(exten_tree, sizeof(exten_tree), "ipcontact/extension/%s", row[0]);
				snprintf(profile_tree, sizeof(profile_tree), "ipcontact/extension/%s/profile", row[0]);
				for (i = 1; i < num_fields; i++) {
					if (row[i]) {
						if (!strcmp(fields[i].name, IPC_PROFILE_LOCAL) ||
							!strcmp(fields[i].name, IPC_PROFILE_CELLULAR) ||
							!strcmp(fields[i].name, IPC_PROFILE_LONGDIST) ||
							!strcmp(fields[i].name, IPC_PROFILE_INTER)) {
							ast_db_put(profile_tree, fields[i].name, row[i]);
						} else {
							ast_db_put(exten_tree, fields[i].name, row[i]);
						}
					}
				}
			}
		} else {
			ast_log(LOG_WARNING,
					"IPContact: Could not find any extensions in database %s@%s\n",
					ipcontact.dbname, ipcontact.dbhost);
		}
		mysql_free_result(sqlres);
		sqlres = NULL;
	} else {
		ast_log(LOG_WARNING,
				"IPContact: Unable to retrieve extensions info from database: %s@%s\n",
				ipcontact.dbname, ipcontact.dbhost);
		res = -1;
	}
	ast_mutex_unlock(&ipc_lock);
	/***** END MUTEX BLOCK **********/
	return res;
}


/*!
  \brief read from MySQL and reload into memory the list of virtual pbxs
  \returns 0 on success or -1 when an error is found
*/
static int reload_vpbxs(void)
{
	MYSQL *conn;
	MYSQL_ROW row;
	MYSQL_RES *sqlres = NULL;
	MYSQL_FIELD *fields = NULL;
	struct virtual_pbx *tmp = NULL, *tmp2 = NULL;
	int i = 0, res = 0;
	int num_fields;

	const char *sql =
		"SELECT number,name,phone,family,mainmenu,operator,dialoptsloc,dialoptsext,dialoptsret,vmslocal,vmsincoming,monitor FROM vpbx ORDER BY number";


	/***** BEGIN MUTEX BLOCK **********/
	ast_mutex_lock(&ipc_lock);

	conn = ipc_sql_reconnect();
	if (!connected) {
		ast_log(LOG_ERROR, "Unable to connect to MySQL server\n");
		res = -1;
		goto out;
	};

	/* Execution. */
	if (mysql_real_query(conn, sql, strlen(sql))) {
		ast_log(LOG_ERROR, "IPContact reload vpbxs: Failed to query database.\n");
		ast_log(LOG_ERROR, "IPContact reload vpbxs: Query: %s\n", sql);
		ast_log(LOG_ERROR, "IPContact reload vpbxs: MySQL error message: %s\n",
				mysql_error(conn));
		res = -1;
		goto out;
	}

	if (!(sqlres = mysql_store_result(conn))) {
		ast_log(LOG_WARNING,
				"IPContact: Unable to retrieve vpbxs info from database: %s@%s\n",
				ipcontact.dbname, ipcontact.dbhost);
		res = -1;
		goto out;
	}
	/* Free current vpbxs list */
	flush_vpbxs();
	AST_LIST_LOCK(&vpbxs);

	if (mysql_num_rows(sqlres) > 0) {
		num_fields = mysql_num_fields(sqlres);
		fields = mysql_fetch_fields(sqlres);
		while ((row = mysql_fetch_row(sqlres))) {
			tmp = (struct virtual_pbx *) malloc(sizeof(struct virtual_pbx));
			if (!tmp) {
				ast_log(LOG_ERROR, "MALLOC FAILED\n");
				mysql_free_result(sqlres);
				sqlres = NULL;
				res = -1;
				AST_LIST_UNLOCK(&vpbxs);
				mysql_free_result(sqlres);
				goto out;
			}
			tmp = make_default_vpbx();
			for (i = 0; i < num_fields; i++) {
				if (row[i] && !ast_strlen_zero(row[i])) {
					if (!strcmp(fields[i].name, "number")) {
						tmp->number = strtol(row[i], (char **) NULL, 10);
						if (debug)
							ast_log(LOG_NOTICE, "Create virtual pbx %d\n", tmp->number);
					} else if (!strcmp(fields[i].name, "phone")) {
						ast_copy_string(tmp->phone, row[i], sizeof(tmp->phone));
					} else if (!strcmp(fields[i].name, "name")) {
						ast_copy_string(tmp->name, row[i], sizeof(tmp->name));
					} else if (!strcmp(fields[i].name, "family")) {
						tmp->family = strtol(row[i], (char **) NULL, 10);
					} else if (!strcmp(fields[i].name, "mainmenu")) {
						ast_copy_string(tmp->mainmenu, row[i], sizeof(tmp->mainmenu));
					} else if (!strcmp(fields[i].name, "operator")) {
						ast_copy_string(tmp->operator, row[i], sizeof(tmp->operator));
					} else if (!strcmp(fields[i].name, "dialoptsloc")) {
						ast_copy_string(tmp->dialoptsloc, row[i], sizeof(tmp->dialoptsloc));
					} else if (!strcmp(fields[i].name, "dialoptsext")) {
						ast_copy_string(tmp->dialoptsext, row[i], sizeof(tmp->dialoptsext));
					} else if (!strcmp(fields[i].name, "dialoptsret")) {
						ast_copy_string(tmp->dialoptsret, row[i], sizeof(tmp->dialoptsret));
					} else if (!strcmp(fields[i].name, "vmslocal")) {
						ast_copy_string(tmp->vmslocal, row[i], sizeof(tmp->vmslocal));
					} else if (!strcmp(fields[i].name, "vmsincoming")) {
						ast_copy_string(tmp->vmsincoming, row[i], sizeof(tmp->vmsincoming));
					} else if (!strcmp(fields[i].name, "monitor")) {
						tmp->monitor = ast_true(row[i]);
					}
				}
			}

			if (AST_LIST_EMPTY(&vpbxs)) {	/* vpbxlist was empty */
				if (tmp->number != 0) {	/* first virtual pbx, must be 0 */
					ast_log(LOG_WARNING,
							"No PBX 0 in DB (first is #%d). Using default values!\n",
							tmp->number);
					if ((tmp2 = make_default_vpbx())) {
						AST_LIST_INSERT_HEAD(&vpbxs, tmp2, list);
						vpbx_count++;
						// Fall through if we have created default pbx
					} else {
						ast_log(LOG_ERROR, "Unable to create default PBX\n");
						res = -1;
						AST_LIST_UNLOCK(&vpbxs);
						mysql_free_result(sqlres);
						goto out;
					}
				}
			}
			if (!AST_LIST_EMPTY(&vpbxs) && (AST_LIST_LAST(&vpbxs)->number == tmp->number)) {
				ast_log(LOG_ERROR,
						"Ignoring duplicated virtual pbx number (%d) entry in DB\n",
						tmp->number);
				free(tmp);
			} else {
				AST_LIST_INSERT_TAIL(&vpbxs, tmp, list);
				vpbx_count++;
			}
		}						// END: while
	} else {
		ast_log(LOG_WARNING,
				"IPContact:  No Virtual PBXs found in MySQL server. Using default VPBX 0\n");
		if ((tmp = make_default_vpbx())) {
			AST_LIST_INSERT_HEAD(&vpbxs, tmp, list);
			vpbx_count++;
		} else {
			ast_log(LOG_ERROR, "Unable to create default PBX\n");
			res = -1;
		}
	}
	AST_LIST_UNLOCK(&vpbxs);

  out:
	sqlres = NULL;
	ast_mutex_unlock(&ipc_lock);
	/***** END MUTEX BLOCK **********/
	if (res == 0)
		astdb_write_vpbxs();
	else
		res = astdb_read_vpbxs();
	return res;
}


/*!
 * \brief Reload the agents info from MYSQL into Asterisk DB
 * \return 0 on success or -1 if unable to connect or query fails
 */
static int reload_agent_info(void)
{
	MYSQL *conn;
	MYSQL_RES *sqlres = NULL;
	MYSQL_FIELD *fields;
	MYSQL_ROW row;
	int num_fields;
	char agent_tree[128];
	char profile_tree[128];
	int i, res = 0;
	char sql[IPC_QUERY_LEN];

	if (schemaversion == 1) {
		snprintf(sql, sizeof(sql),
						 "select u.username,local,longdist,cellular,inter from authcalls a, agents ag, users u where a.username=ag.user and a.username=u.username and disabled=0");
	} else {
		snprintf(sql, sizeof(sql),
						 "select username,local,longdist,cellular,inter from authcalls a, agents ag, users u where a.user_id=ag.user_id and a.user_id=u.user_id and disabled=0");
	}

	/***** BEGIN MUTEX BLOCK **********/
	ast_mutex_lock(&ipc_lock);

	conn = ipc_sql_reconnect();
	if (!conn) {
		ast_log(LOG_ERROR, "Unable to connect to MySQL server\n");
		ast_mutex_unlock(&ipc_lock);
		return -1;
	};

	/* Execution. */
	if (mysql_real_query(conn, sql, strlen(sql))) {
		ast_log(LOG_ERROR, "IPContact reload agents: Failed to query database.\n");
		ast_log(LOG_ERROR, "IPContact reload agents: Query: %s\n", sql);
		ast_log(LOG_ERROR, "IPContact reload agents: MySQL error message: %s\n",
				mysql_error(conn));
		ast_mutex_unlock(&ipc_lock);
		return -1;
	}

	if ((sqlres = mysql_store_result(conn))) {
		/* Borro todo el arbol de agentes antes de empezar siempre */
		ast_db_deltree("ipcontact/agent", 0);

		if (mysql_num_rows(sqlres) > 0) {
			num_fields = mysql_num_fields(sqlres);
			fields = mysql_fetch_fields(sqlres);

			while ((row = mysql_fetch_row(sqlres))) {
				snprintf(agent_tree, sizeof(agent_tree), "ipcontact/agent/%s", row[0]);
				snprintf(profile_tree, sizeof(profile_tree), "ipcontact/agent/%s/profile", row[0]);
				for (i = 1; i < num_fields; i++) {
					if (row[i]) {
						if (!strcmp(fields[i].name, IPC_PROFILE_LOCAL) ||
							!strcmp(fields[i].name, IPC_PROFILE_CELLULAR) ||
							!strcmp(fields[i].name, IPC_PROFILE_LONGDIST) ||
							!strcmp(fields[i].name, IPC_PROFILE_INTER)) {
							ast_db_put(profile_tree, fields[i].name, row[i]);
						} else {
							ast_db_put(agent_tree, fields[i].name, row[i]);
						}
					}
				}
			}
			if (debug)
				ast_log(LOG_NOTICE, "Reloading agent profiles into Asterisk DB\n");
		} else {
			ast_log(LOG_NOTICE,
					"IPContact: Could not find any agents with profiles in database %s@%s\n",
					ipcontact.dbname, ipcontact.dbhost);
		}
		mysql_free_result(sqlres);
		sqlres = NULL;
	} else {
		ast_log(LOG_WARNING,
				"IPContact: Unable to retrieve agents info from database: %s@%s\n",
				ipcontact.dbname, ipcontact.dbhost);
		res = -1;
	}
	ast_mutex_unlock(&ipc_lock);
	/***** END MUTEX BLOCK **********/
	return res;
}


/*!
 * \brief Reload the agents info from MYSQL into Asterisk DB
 * \return 0 on success or -1 if unable to connect or query fails
 */
static int reload_tables_info(void)
{
	MYSQL *conn;
	MYSQL_RES *sqlres = NULL;
	MYSQL_FIELD *fields;
	MYSQL_ROW row;
	int num_fields;
	char table_tree[128];
	int i, res = 0;

	char sql[IPC_MAX_LEN];
	struct ast_ipcontact_table *cur = NULL;	

	/***** BEGIN MUTEX BLOCK **********/
	ast_mutex_lock(&ipc_lock);

	conn = ipc_sql_reconnect();
	if (!conn) {
		ast_log(LOG_ERROR, "Unable to connect to MySQL server\n");
		ast_mutex_unlock(&ipc_lock);
		return -1;
	};

	/* Borro todo el arbol de agentes antes de empezar siempre */
	ast_db_deltree("ipcontact/table", 0);

	/* Iterate over each custom table */
	AST_LIST_TRAVERSE(&tables, cur, list) {
		if (debug)
			ast_log(LOG_NOTICE, "Reloading custom table %s into Asterisk DB\n", cur->table);
		snprintf(sql, sizeof(sql), "select * from %s", cur->table);

		/* Execution. */
		if (mysql_real_query(conn, sql, strlen(sql))) {
			ast_log(LOG_ERROR, "IPContact reload tables: Failed to query database.\n");
			ast_log(LOG_ERROR, "IPContact reload tables: Query: %s\n", sql);
			ast_log(LOG_ERROR, "IPContact reload tables: MySQL error message: %s\n",
					mysql_error(conn));
			ast_mutex_unlock(&ipc_lock);
			return -1;
		}

		if ((sqlres = mysql_store_result(conn))) {
			if (mysql_num_rows(sqlres) > 0) {
				num_fields = mysql_num_fields(sqlres);
				fields = mysql_fetch_fields(sqlres);

				while ((row = mysql_fetch_row(sqlres))) {
					snprintf(table_tree, sizeof(table_tree), "ipcontact/table/%s/%s", cur->table, row[0]);
					for (i = 1; i < num_fields; i++) {
						if (row[i])
							ast_db_put(table_tree, fields[i].name, row[i]);
					}
				}
			} else {
				ast_log(LOG_WARNING,
						"IPContact: Could not find any rows in custom table %s, database %s@%s\n",
						cur->table, ipcontact.dbname, ipcontact.dbhost);
			}
			mysql_free_result(sqlres);
			sqlres = NULL;
		} else {
			ast_log(LOG_WARNING,
					"IPContact: Unable to retrieve custom tables info from database: %s@%s\n",
					ipcontact.dbname, ipcontact.dbhost);
			res = -1;
		}
	}

	
	ast_mutex_unlock(&ipc_lock);
	/***** END MUTEX BLOCK **********/
	return res;
}


struct virtual_pbx *get_virtual_pbx(int vpbx)
{
	struct virtual_pbx *cur = NULL;
	AST_LIST_TRAVERSE(&vpbxs, cur, list) {
		if (cur->number == vpbx) {
			break;
		}
	}
	return cur;
}


/*! \brief Write the in-memory list of virtual pbxs into Asterisk DB */
static void astdb_write_vpbxs(void)
{

	struct virtual_pbx *cur;
	char vpbx_tree[IPC_MAX_LEN];
	char buffer[16];

	ast_db_deltree("ipcontact/vpbx", 0);

	AST_LIST_LOCK(&vpbxs);
	AST_LIST_TRAVERSE(&vpbxs, cur, list) {
		snprintf(vpbx_tree, sizeof(vpbx_tree), "ipcontact/vpbx/%d", cur->number);
		ast_db_put(vpbx_tree, "name", cur->name);
		ast_db_put(vpbx_tree, "phone", cur->phone);
		snprintf(buffer, sizeof(buffer), "%d", cur->family);
		ast_db_put(vpbx_tree, "family", buffer);
		ast_db_put(vpbx_tree, "mainmenu", cur->mainmenu);
		ast_db_put(vpbx_tree, "operator", cur->operator);
		ast_db_put(vpbx_tree, "vmsincoming", cur->vmsincoming);
		ast_db_put(vpbx_tree, "vmslocal", cur->vmslocal);
		ast_db_put(vpbx_tree, "dialoptsloc", cur->dialoptsloc);
		ast_db_put(vpbx_tree, "dialoptsext", cur->dialoptsext);
		ast_db_put(vpbx_tree, "dialoptsret", cur->dialoptsret);
		snprintf(buffer, sizeof(buffer), "%d", cur->monitor);
		ast_db_put(vpbx_tree, "monitor", buffer);
	}
	AST_LIST_UNLOCK(&vpbxs);
}


/*! \brief Read Asterisk DB and fill the in-memory list of virtual pbxs */
static int astdb_read_vpbxs(void)
{

	static struct ast_db_entry *dbe;
	static struct virtual_pbx *cur;
	int res = 0;
	int vpbxnum = 0;
	char vpbxnum_str[3];

	/***** BEGIN MUTEX BLOCK **********/
	ast_mutex_lock(&ipc_lock);
	ast_log(LOG_WARNING, "Failsafe mode: reloading VPBX list from Asterisk DB!\n");
	flush_vpbxs();
	AST_LIST_LOCK(&vpbxs);
	snprintf(vpbxnum_str, sizeof(vpbxnum_str), "%i", vpbxnum);
	while ((dbe = ast_db_gettree("ipcontact/vpbx", vpbxnum_str))) {
		cur = (struct virtual_pbx *) malloc(sizeof(struct virtual_pbx));
		if (!cur) {
			ast_log(LOG_ERROR, "MALLOC FAILED\n");
			res = -1;
			break;
		}
		cur = make_default_vpbx();
		cur->number = vpbxnum;
		if (debug)
			ast_log(LOG_NOTICE, "Read virtual pbx %d from Asterisk DB...\n", vpbxnum);
		for (; dbe ; dbe = dbe->next) {
			if (strstr(dbe->key, "phone")) {
				ast_copy_string(cur->phone, dbe->data, sizeof(cur->phone));
			} else if (strstr(dbe->key, "name")) {
				ast_copy_string(cur->name, dbe->data, sizeof(cur->name));
			} else if (strstr(dbe->key, "family")) {
				cur->family = strtol(dbe->data, (char **) NULL, 10);
			} else if (strstr(dbe->key, "mainmenu")) {
				ast_copy_string(cur->mainmenu, dbe->data, sizeof(cur->mainmenu));
			} else if (strstr(dbe->key, "operator")) {
				ast_copy_string(cur->operator, dbe->data, sizeof(cur->operator));
			} else if (strstr(dbe->key, "dialoptsloc")) {
				ast_copy_string(cur->dialoptsloc, dbe->data, sizeof(cur->dialoptsloc));
			} else if (strstr(dbe->key, "dialoptsext")) {
				ast_copy_string(cur->dialoptsext, dbe->data, sizeof(cur->dialoptsext));
			} else if (strstr(dbe->key, "dialoptsret")) {
				ast_copy_string(cur->dialoptsret, dbe->data, sizeof(cur->dialoptsret));
			} else if (strstr(dbe->key, "vmslocal")) {
				ast_copy_string(cur->vmslocal, dbe->data, sizeof(cur->vmslocal));
			} else if (strstr(dbe->key, "vmsincoming")) {
				ast_copy_string(cur->vmsincoming, dbe->data, sizeof(cur->vmsincoming));
			} else if (strstr(dbe->key, "monitor")) {
				cur->monitor = ast_true(dbe->data);
			} else {
				ast_log(LOG_WARNING, "Invalid key %s: %s. Ignoring it\n", dbe->key, dbe->data);
			}
		}
		AST_LIST_INSERT_TAIL(&vpbxs, cur, list);
		if (debug)
			ast_log(LOG_NOTICE, "Create virtual pbx %d\n", cur->number);
		vpbxnum++;
		snprintf(vpbxnum_str, sizeof(vpbxnum_str), "%i", vpbxnum);
	}
	if (AST_LIST_EMPTY(&vpbxs))
		ast_log(LOG_WARNING, "No virtual pbx's found in Asterisk DB! Good luck...\n");
	
	AST_LIST_UNLOCK(&vpbxs);
	ast_mutex_unlock(&ipc_lock);
	/***** END MUTEX BLOCK **********/

	return res;
}


/*
 * Load module stuff
 */
static int load_config(void)
{

	struct ast_config *config;
	struct ast_variable *var;
	char *cat = NULL;
	char *tab = NULL;
	int res = 0;
	char config_path[PATH_MAX] = "";
	char buffer[IPC_MAX_LEN] = "";
	char *cursor = buffer;
	size_t left = sizeof(buffer);
	snprintf(config_path, sizeof(config_path), "%s/%s", IPCONTACT_CONFDIR, config_file);
	config_path[sizeof(config_path) - 1] = 0;
	struct ast_ipcontact_table *cur;
	const char delim[3] = ",";	
	
#if defined(HAS_AST16) || defined(HAS_AST18)
	struct ast_flags config_flags = { CONFIG_FLAG_WITHCOMMENTS };
	config = ast_config_load((char *) config_path, config_flags);
#else
	config = ast_config_load((char *) config_path);
#endif
	if (!config) {
		ast_log(LOG_ERROR,"Unable to load config file ipcontact.conf for IPContact applications\n");
		return -1;
	};

	ast_mutex_lock(&ipc_lock);

	cat = NULL;
	while ((cat = ast_category_browse(config, cat))) {
		if (!strcasecmp(cat, "general")) {
			for (var = ast_variable_browse(config, cat); var; var = var->next) {
				if (!strcasecmp(var->name, "debug")) {
					if (ast_true(var->value)) {
						ast_log(LOG_NOTICE,"res_ipcontact loaded in debug mode\n");
						debug = 1;
					}
				} else if (!strcasecmp(var->name, "schemaversion")) {
					if (sscanf(var->value, "%d", &schemaversion) == 1) {
						ast_log(LOG_NOTICE, "Using schema version %d\n", schemaversion);
					}
					if (schemaversion != 1 && schemaversion != 2)
						ast_log(LOG_WARNING, "Schema version %d not supported!\n", schemaversion);

				} else {
					res = ast_build_string(&cursor, &left, "IPCONTACT_%s", var->name);
					if (res) {
						ast_log(LOG_ERROR,"Unable to define ENV variable: %s\n", var->name);
					} else {
						setenv(buffer, var->value, 1);
						ast_log(LOG_NOTICE,"Defined ENV variable: %s=%s\n", buffer,
								var->value);
					}
					cursor = buffer;
					left = sizeof(buffer);
				}
			}
		} else if (!strcasecmp(cat, "sql")) {
			flush_tables();
			for (var = ast_variable_browse(config, cat); var; var = var->next) {
				if (!strcasecmp(var->name, "host")) {
					ast_copy_string(ipcontact.dbhost, var->value, sizeof(ipcontact.dbhost));
				} else if (!strcasecmp(var->name, "name")) {
					ast_copy_string(ipcontact.dbname, var->value, sizeof(ipcontact.dbname));
				} else if (!strcasecmp(var->name, "user")) {
					ast_copy_string(ipcontact.dbuser, var->value, sizeof(ipcontact.dbuser));
				} else if (!strcasecmp(var->name, "pass")) {
					ast_copy_string(ipcontact.dbpass, var->value, sizeof(ipcontact.dbpass));
				} else if (!strcasecmp(var->name, "port")) {
					if ((sscanf(var->value, "%d", &ipcontact.dbport) != 1) || ipcontact.dbport < 0 || ipcontact.dbport > 65535) {
						ast_log(LOG_ERROR, "Bad port number %s\n", var->value);
						ipcontact.dbport = 3306;
					}
				} else if (!strcasecmp(var->name, "sock")) {
					ast_copy_string(ipcontact.dbsock, var->value, sizeof(ipcontact.dbsock));
				} else if (!strcasecmp(var->name, "tables")) {
					ast_copy_string(ipcontact.tables, var->value, sizeof(ipcontact.tables));
					/* Parse custom tables string */
					AST_LIST_LOCK(&tables);
#ifdef HAS_AST18
					while ((tab = strsep((char **) &var->value, delim))) {
#else
					while ((tab = strsep(&var->value, delim))) {
#endif
						cur = (struct ast_ipcontact_table *) malloc(sizeof(struct ast_ipcontact_table));
						memset(cur, 0, sizeof(struct ast_ipcontact_table));
						if (!cur) {
							ast_log(LOG_ERROR, "MALLOC FAILED\n");
							res = -1;
							AST_LIST_UNLOCK(&tables);
							goto fail;
						}
						ast_copy_string(cur->table, ast_strip(tab), sizeof(cur->table));
						if (strlen(cur->table) > 0) {
							AST_LIST_INSERT_TAIL(&tables, cur, list);
							/* ast_log(LOG_WARNING, "tabla %s count %i\n", cur->table, table_count); */
							table_count++;
						}
					}
					AST_LIST_UNLOCK(&tables);					
				}
			}
			/* Force a db reconnect */
			connected = 0;
		} else if (!strcasecmp(cat, "vpbx")) {
			for (var = ast_variable_browse(config, cat); var; var = var->next) {
				if (!strcasecmp(var->name, "mainmenu")) {
					ast_copy_string(default_vpbx.mainmenu, var->value, sizeof(default_vpbx.mainmenu));
				} else if (!strcasecmp(var->name, "operator")) {
					ast_copy_string(default_vpbx.operator, var->value, sizeof(default_vpbx.operator));
				} else if (!strcasecmp(var->name, "vmslocal")) {
					ast_copy_string(default_vpbx.vmslocal, var->value, sizeof(default_vpbx.vmslocal));
				} else if (!strcasecmp(var->name, "vmsincoming")) {
					ast_copy_string(default_vpbx.vmsincoming, var->value, sizeof(default_vpbx.vmsincoming));
				} else if (!strcasecmp(var->name, "dialoptsloc")) {
					ast_copy_string(default_vpbx.dialoptsloc, var->value, sizeof(default_vpbx.dialoptsloc));
				} else if (!strcasecmp(var->name, "dialoptsext")) {
					ast_copy_string(default_vpbx.dialoptsext, var->value, sizeof(default_vpbx.dialoptsext));
				} else if (!strcasecmp(var->name, "dialoptsret")) {
					ast_copy_string(default_vpbx.dialoptsret, var->value, sizeof(default_vpbx.dialoptsret));
				} else if (!strcasecmp(var->name, "monitor")) {
					default_vpbx.monitor = ast_true(var->value) ? 1:0;
				}
			}
		} else {
			if (debug) {
				ast_verbose("Parsing config context: [%s]\n", cat);
			}
			snprintf(buffer, sizeof(buffer), "%s/%s", db_root, cat);
			ast_db_deltree(buffer, 0);
			for (var = ast_variable_browse(config, cat); var; var = var->next) {
				if (var->value && (strlen(var->value) > 0)) {
					if ((res = ast_db_put(buffer, var->name, var->value))) {
						ast_log(LOG_ERROR,
								"Unable to set AstDB entry: family=%s, key=%s, value=%s.\n",
								buffer, var->name, var->value);
						goto fail;
					} else if (debug) {
						ast_verbose("%s/%s = %s\n", buffer, var->name, var->value);
					}
				} else {
					ast_log(LOG_WARNING, "Invalid empty value for family=%s key=%s\n",
							cat, var->name);
				}
			}
		}
	}
	/* Convert null fields to default values */
	if (!strcmp(ipcontact.dbhost, "")) {
		ast_log(LOG_WARNING,
				"MySQL server hostname not specified. Using \"localhost\"\n");
		strcpy(ipcontact.dbhost, "localhost");
	}
	if (!strcmp(ipcontact.dbname, "")) {
		ast_log(LOG_WARNING, "MySQL database not specified. Using \"ipcontact\"\n");
		strcpy(ipcontact.dbname, "ipcontact");
	}
	if (!strcmp(ipcontact.dbuser, "")) {
		ast_log(LOG_WARNING, "MySQL database user not specified. Using \"root\"\n");
		strcpy(ipcontact.dbuser, "root");
	}
	if (AST_LIST_EMPTY(&tables)) {
		ast_log(LOG_NOTICE, "No custom tables configured\n");
	}

	if (debug) {
		ast_verbose("IPContact DB: host=%s, name=%s, user=%s, pass=%s, port=%d, sock=%s\n",
					ipcontact.dbhost, ipcontact.dbname, ipcontact.dbuser,
					ipcontact.dbpass, ipcontact.dbport, ipcontact.dbsock);

		if (!AST_LIST_EMPTY(&tables)) {
			ast_verbose("Using %i custom tables: ", table_count);
			AST_LIST_TRAVERSE(&tables, cur, list) {
				ast_verbose("%s ", cur->table);
			}
			ast_verbose("\n");
		}
	}
	
	
  fail:
	/* Liberamos lock ipc_lock */
	ast_mutex_unlock(&ipc_lock);
	ast_config_destroy(config);
	return res;
};


/*
 * IPCONTACT RELOAD CONFIG command handler
 */
#ifdef HAS_AST18
static int ipcontact_reload_config(int fd, int argc, const char * const argv[])
#else
static int ipcontact_reload_config(int fd, int argc, char *argv[])
#endif
{
	if (argc != 3)
		return RESULT_SHOWUSAGE;
	if (load_config())
		return -1;
	return 0;
};

#if defined(HAS_AST16) || defined(HAS_AST18)
static char *handle_ipcontact_reload_config(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	switch ( cmd ) {
	case CLI_INIT:
		e->command = "ipcontact reload config";
		e->usage = help_reload_config;
		return NULL;
	case CLI_GENERATE:
		return NULL;
	}
	
	if (ipcontact_reload_config(a->fd, a->argc, a->argv))
		return CLI_SHOWUSAGE;
	else
		return CLI_SUCCESS;

}
#endif

/*
 * IPCONTACT RELOAD EXTENSIONS command handler
 */
#ifdef HAS_AST18
static int ipcontact_reload_extensions(int fd, int argc, const char * const argv[])
#else
static int ipcontact_reload_extensions(int fd, int argc, char *argv[])
#endif
{
	if (argc != 3)
		return RESULT_SHOWUSAGE;
	if (reload_exten_info())
		return -1;
	return 0;
};

#if defined(HAS_AST16) || defined(HAS_AST18)
static char *handle_ipcontact_reload_extensions(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	switch ( cmd ) {
	case CLI_INIT:
		e->command = "ipcontact reload extensions";
		e->usage = help_reload_extensions;
		return NULL;
	case CLI_GENERATE:
		return NULL;
	}
	if (ipcontact_reload_extensions(a->fd, a->argc, a->argv))
		return CLI_SHOWUSAGE;
	else
		return CLI_SUCCESS;
}	
#endif

/*
 * IPCONTACT RELOAD VPBXS command handler
 */
#ifdef HAS_AST18
static int ipcontact_reload_vpbxs(int fd, int argc, const char * const argv[])
#else
static int ipcontact_reload_vpbxs(int fd, int argc, char *argv[])
#endif
{
	if (argc != 3)
		return RESULT_SHOWUSAGE;
	if (reload_vpbxs())
		return -1;
	return 0;
};

#if defined(HAS_AST16) || defined(HAS_AST18)
static char *handle_ipcontact_reload_vpbxs(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	switch ( cmd ) {
	case CLI_INIT:
		e->command = "ipcontact reload vpbxs";
		e->usage = help_reload_vpbxs;
		return NULL;
	case CLI_GENERATE:
		return NULL;
	}
	if (ipcontact_reload_vpbxs(a->fd, a->argc, a->argv))
		return CLI_SHOWUSAGE;
	else
		return CLI_SUCCESS;
}	
#endif

/*
 * IPCONTACT RELOAD AGENTS command handler
 */
#ifdef HAS_AST18
static int ipcontact_reload_agents(int fd, int argc, const char * const argv[])
#else
static int ipcontact_reload_agents(int fd, int argc, char *argv[])
#endif
{
	if (argc != 3)
		return RESULT_SHOWUSAGE;
	if (reload_agent_info())
		return -1;
	return 0;
};

#if defined(HAS_AST16) || defined(HAS_AST18)
static char *handle_ipcontact_reload_agents(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	switch ( cmd ) {
	case CLI_INIT:
		e->command = "ipcontact reload agents";
		e->usage = help_reload_agents;
		return NULL;
	case CLI_GENERATE:
		return NULL;
	}
	if (ipcontact_reload_agents(a->fd, a->argc, a->argv))
		return CLI_SHOWUSAGE;
	else
		return CLI_SUCCESS;
}	
#endif

/*
 * IPCONTACT RELOAD TABLES command handler
 */
#ifdef HAS_AST18
static int ipcontact_reload_tables(int fd, int argc, const char * const argv[])
#else
static int ipcontact_reload_tables(int fd, int argc, char *argv[])
#endif
{
	if (argc != 3)
		return RESULT_SHOWUSAGE;
	if (reload_tables_info())
		return -1;
	return 0;
};

#if defined(HAS_AST16) || defined(HAS_AST18)
static char *handle_ipcontact_reload_tables(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	switch ( cmd ) {
	case CLI_INIT:
		e->command = "ipcontact reload tables";
		e->usage = help_reload_tables;
		return NULL;
	case CLI_GENERATE:
		return NULL;
	}
	if (ipcontact_reload_tables(a->fd, a->argc, a->argv))
		return CLI_SHOWUSAGE;
	else
		return CLI_SUCCESS;
}	
#endif

/*
 * IPCONTACT SHOW VPBXS command handler
 */
#ifdef HAS_AST18
static int ipcontact_show_vpbxs(int fd, int argc, const char * const argv[])
#else
static int ipcontact_show_vpbxs(int fd, int argc, char *argv[])
#endif
{
#define FORMAT  "  %-14.14s:\t%s\n"
#define FORMAT2  "  %-14.14s:\t%d\n"
	char output[1024], *cursor;
	struct virtual_pbx *cur;

	if (argc != 3)
		return RESULT_SHOWUSAGE;
	if (AST_LIST_EMPTY(&vpbxs)) {
		ast_log(LOG_WARNING, "VPBX list is empty! Check DB connection\n");
	} else {
		AST_LIST_LOCK(&vpbxs);
		AST_LIST_TRAVERSE(&vpbxs, cur, list) {
			cursor = output;
			cursor += snprintf(output, sizeof(output), "* Virtual PBX %d\n", cur->number);
			cursor +=
				snprintf(cursor, sizeof(output) - (cursor - output), FORMAT, "Name",
						 cur->name);
			cursor +=
				snprintf(cursor, sizeof(output) - (cursor - output), FORMAT2, "Family",
						 cur->family);
			cursor +=
				snprintf(cursor, sizeof(output) - (cursor - output), FORMAT,
						 "Phone Number", cur->phone);
			cursor +=
				snprintf(cursor, sizeof(output) - (cursor - output), FORMAT, "Main Menu",
						 cur->mainmenu);
			cursor +=
				snprintf(cursor, sizeof(output) - (cursor - output), FORMAT, "Operator",
						 cur->operator);
			cursor +=
				snprintf(cursor, sizeof(output) - (cursor - output), FORMAT,
						 "Monitor Calls", cur->monitor ? "yes" : "no");
			cursor +=
				snprintf(cursor, sizeof(output) - (cursor - output), "  Dial Options\n");
			cursor +=
				snprintf(cursor, sizeof(output) - (cursor - output), FORMAT, "- local",
						 cur->dialoptsloc);
			cursor +=
				snprintf(cursor, sizeof(output) - (cursor - output), FORMAT, "- extern",
						 cur->dialoptsext);
			cursor +=
				snprintf(cursor, sizeof(output) - (cursor - output), FORMAT, "- return",
						 cur->dialoptsret);
			cursor +=
				snprintf(cursor, sizeof(output) - (cursor - output),
						 "  Voicemail Context\n");
			cursor +=
				snprintf(cursor, sizeof(output) - (cursor - output), FORMAT, "- Incoming",
						 cur->vmsincoming);
			cursor +=
				snprintf(cursor, sizeof(output) - (cursor - output), FORMAT, "- Local",
						 cur->vmslocal);
			ast_cli(fd, "%s\n", output);
		}
		AST_LIST_UNLOCK(&vpbxs);
	}

	return RESULT_SUCCESS;
#undef FORMAT
}

#if defined(HAS_AST16) || defined(HAS_AST18)
static char *handle_ipcontact_show_vpbxs(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	switch ( cmd ) {
	case CLI_INIT:
		e->command = "ipcontact show vpbxs";
		e->usage = help_show_vpbxs;
		return NULL;
	case CLI_GENERATE:
		return NULL;
	}
	if (ipcontact_show_vpbxs(a->fd, a->argc, a->argv))
		return CLI_SHOWUSAGE;
	else
		return CLI_SUCCESS;
}	
#endif

/*
 * IPCONTACT SHOW VERSION command handler
 */
#ifdef HAS_AST18
static int ipcontact_show_version(int fd, int argc, const char * const argv[])
#else
static int ipcontact_show_version(int fd, int argc, char *argv[])
#endif
{
	if (argc != 3)
		return RESULT_SHOWUSAGE;
	ast_cli(fd, "IPContact Addons " IPCONTACT_ADDONS_VERSION "\n");
	return RESULT_SUCCESS;
}

#if defined(HAS_AST16) || defined(HAS_AST18)
static char *handle_ipcontact_show_version(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	switch ( cmd ) {
	case CLI_INIT:
		e->command = "ipcontact show version";
		e->usage = help_show_version;
		return NULL;
	case CLI_GENERATE:
		return NULL;
	}
	if (ipcontact_show_version(a->fd, a->argc, a->argv))
		return CLI_SHOWUSAGE;
	else
		return CLI_SUCCESS;
}	
#endif


/*
 * IPCONTACT SHOW CONFIG command handler
 */
#ifdef HAS_AST18
static int ipcontact_show_config(int fd, int argc, const char * const argv[])
#else
static int ipcontact_show_config(int fd, int argc, char *argv[])
#endif
{
	char output[1024];
	if (argc != 3)
		return RESULT_SHOWUSAGE;
	ast_cli(fd, "IPContact Addons Configuration Parameters (ipcontact.conf)\n\n");
	snprintf(output, sizeof(output),
			 "* General\n"
			 "  schemaversion : %d\n"
			 "  debug         : %s\n"
			 "\n* SQL\n"
			 "  dbhost        : %s\n"
			 "  dbname        : %s\n"
			 "  dbuser        : %s\n"
			 "  dbpass        : %s\n"
			 "  dbport        : %d\n"
			 "  dbsock        : %s\n"
			 "  tables        : %s\n",
			 schemaversion, debug ? "yes" : "no",
			 ipcontact.dbhost, ipcontact.dbname, ipcontact.dbuser, ipcontact.dbpass, ipcontact.dbport, ipcontact.dbsock, ipcontact.tables);

	ast_cli(fd, "%s\n", output);		
	return RESULT_SUCCESS;
}

#if defined(HAS_AST16) || defined(HAS_AST18)
static char *handle_ipcontact_show_config(struct ast_cli_entry *e, int cmd, struct ast_cli_args *a)
{
	switch ( cmd ) {
	case CLI_INIT:
		e->command = "ipcontact show config";
		e->usage = help_show_config;
		return NULL;
	case CLI_GENERATE:
		return NULL;
	}
	if (ipcontact_show_config(a->fd, a->argc, a->argv))
		return CLI_SHOWUSAGE;
	else
		return CLI_SUCCESS;
}	
#endif


/*
 * CLI entries for commands provided by this module
 */
#if defined(HAS_AST16) || defined(HAS_AST18)
static struct ast_cli_entry cli_ipcontact[] = {
	AST_CLI_DEFINE(handle_ipcontact_reload_config, "Reload IPContact configuration"),
	AST_CLI_DEFINE(handle_ipcontact_reload_extensions, "Reload IPContact Extensions info into AstDB"),
	AST_CLI_DEFINE(handle_ipcontact_reload_vpbxs, "Reload IPContact Virtual PBX info into AstDB"),
	AST_CLI_DEFINE(handle_ipcontact_reload_agents, "Reload IPContact Agents info into AstDB"),
	AST_CLI_DEFINE(handle_ipcontact_reload_tables, "Reload IPContact Custom Tables info into AstDB"),
	AST_CLI_DEFINE(handle_ipcontact_show_vpbxs, "Show IPContact Virtual PBX list"),
	AST_CLI_DEFINE(handle_ipcontact_show_version, "Show IPContact Addons Version"),
	AST_CLI_DEFINE(handle_ipcontact_show_config, "Show IPContact Addons Config"),
};
#else
static struct ast_cli_entry cli_ipcontact[] = {
	{{"ipcontact", "reload", "config", NULL},
	 ipcontact_reload_config, "Reload IPContact configuration",
	 help_reload_config, NULL},
	{{"ipcontact", "reload", "extensions", NULL},
	 ipcontact_reload_extensions, "Reload IPContact Extensions info into AstDB",
	 help_reload_extensions, NULL},
	{{"ipcontact", "reload", "vpbxs", NULL},
	 ipcontact_reload_vpbxs, "Reload IPContact Virtual PBX info into AstDB",
	 help_reload_vpbxs, NULL},
	{{"ipcontact", "reload", "agents", NULL},
	 ipcontact_reload_agents, "Reload IPContact Agents info into AstDB",
	 help_reload_agents, NULL},
	{{"ipcontact", "reload", "tables", NULL},
	 ipcontact_reload_tables, "Reload IPContact Custom Tables info into AstDB",
	 help_reload_tables, NULL},
	{{"ipcontact", "show", "vpbxs", NULL},
	 ipcontact_show_vpbxs, "Show IPContact Virtual PBX list",
	 help_show_vpbxs, NULL},
	{{"ipcontact", "show", "version", NULL},
	 ipcontact_show_version, "Show IPContact Addons version",
	 help_show_version, NULL},
	{{"ipcontact", "show", "config", NULL},
	 ipcontact_show_config, "Show IPContact Addons Config",
	 help_show_config, NULL},
};
#endif


/*
 * Standard module functions ...
 */
static int unload_module(void)
{
	int res = 0;
	ast_mutex_lock(&ipc_lock);
	/* remove the registed structures, CLI entries and functions */
	flush_vpbxs();
	flush_tables();
	ast_cli_unregister_multiple(cli_ipcontact,
								sizeof(cli_ipcontact) / sizeof(struct ast_cli_entry));
	res |= ast_custom_function_unregister(&normantel_function);
	res |= ast_custom_function_unregister(&profileauth_function);
	res |= ast_custom_function_unregister(&extenpin_function);
	res |= ast_custom_function_unregister(&userpin_function);
	res |= ast_custom_function_unregister(&cellexten_function);
	res |= ast_custom_function_unregister(&calleridcellexten_function);
	res |= ast_custom_function_unregister(&ipcmysql_function);
	ipc_sql_disconnect();
	ast_mutex_unlock(&ipc_lock);
	return 0;
};

static int load_module(void)
{
	int res = 0;

	/* Solo doy LOAD_DECLINE si no puedo levantar el archivo de configuracion
	 * Si no me conecto a la base de datos igual doy LOAD_SUCCESS y registro todo */
	if (load_config())
		return AST_MODULE_LOAD_DECLINE;

	res = reload_vpbxs();
	res |= reload_exten_info();
	res |= reload_agent_info();
	res |= reload_tables_info();
	
	ast_custom_function_register(&normantel_function);
	ast_custom_function_register(&profileauth_function);
	ast_custom_function_register(&extenpin_function);
	ast_custom_function_register(&userpin_function);
	ast_custom_function_register(&cellexten_function);
	ast_custom_function_register(&calleridcellexten_function);
	ast_custom_function_register(&ipcmysql_function);
	ast_cli_register_multiple(cli_ipcontact,
							  sizeof(cli_ipcontact) / sizeof(struct ast_cli_entry));
	
	/* Ojo que res=-1 es AST_MODULE_LOAD FAILURE, si retorno res y no me conecto a la base de datos no carga el modulo */
	/* Siempre retorno AST_MODULE_LOAD_SUCCESS independiente de lo que retornen las subfunciones */
	return AST_MODULE_LOAD_SUCCESS;
}

static int reload(void)
{
	int res = 0;
	ipc_sql_disconnect();
	if (load_config())
		return AST_MODULE_LOAD_DECLINE;

	if (!ipc_sql_reconnect()) {
		ast_log(LOG_ERROR, "MySQL Login failed\n");
		res = 1;
	} else {
		res = reload_vpbxs();
		res |= reload_exten_info();
		res |= reload_agent_info();
		res |= reload_tables_info();
	}
	return AST_MODULE_LOAD_SUCCESS;
};


AST_MODULE_INFO(ASTERISK_GPL_KEY, AST_MODFLAG_GLOBAL_SYMBOLS, "IPContact",
								.load = load_module,
								.unload = unload_module,
								.reload = reload,
								);
